/*
 * Copyright © 2013-2014 Collabora, Ltd.
 * Copyright © 2016 DENSO CORPORATION
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "waltham-object.h"
#include "waltham-connection.h"
#include "waltham-private.h"

#include "message.h"
#include "demarshaller.h"

//#define ENABLE_PROFILING 1

// FIXME: we no longer use glib, so this is broken
#ifdef ENABLE_PROFILING
#define START_TIMING() \
  gdouble msec; \
  GTimer *timer = g_timer_new ();
#define END_TIMING(label) \
  msec = g_timer_elapsed (timer, NULL) * 1000; \
  g_timer_destroy (timer); \
  g_debug ("%s (%f ms)", label, msec);
#else
#define START_TIMING()
#define END_TIMING(a)
#endif

#define PADDED(sz) \
   (((sz) + 3) & ~3)

#include "waltham-client.h"

/* Function: wth_display_send_error
 * Opcode 4 (hexadecimal: 0x4)
 */
static void function_4(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wth_display *wth_display = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wth_object *object_id = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));
  uint32_t *code = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  unsigned int message_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  const char *message = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int));

  wth_debug ("wth_display_send_error(%u, %u, %u, [variable type const char *]) (opcode 4) called.", ((struct wth_object *)wth_display)->id, ((struct wth_object *)object_id)->id, *code);
  START_TIMING();
  ((struct wth_display_listener *)(((struct wth_object *)wth_display)->vfunc))->error
    (wth_display, object_id, *code, message);
  END_TIMING("wth_display_send_error");
}

/* Function: wth_display_send_delete_id
 * Opcode 5 (hexadecimal: 0x5)
 */
static void function_5(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wth_display *wth_display = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *id = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wth_display_send_delete_id(%u, %u) (opcode 5) called.", ((struct wth_object *)wth_display)->id, *id);
  START_TIMING();
  ((struct wth_display_listener *)(((struct wth_object *)wth_display)->vfunc))->delete_id
    (wth_display, *id);
  END_TIMING("wth_display_send_delete_id");
}

/* Function: wth_display_send_server_version
 * Opcode 6 (hexadecimal: 0x6)
 */
static void function_6(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wth_display *wth_display = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *server_version = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wth_display_send_server_version(%u, %u) (opcode 6) called.", ((struct wth_object *)wth_display)->id, *server_version);
  START_TIMING();
  ((struct wth_display_listener *)(((struct wth_object *)wth_display)->vfunc))->server_version
    (wth_display, *server_version);
  END_TIMING("wth_display_send_server_version");
}

/* Function: wthp_registry_send_global
 * Opcode 9 (hexadecimal: 0x9)
 */
static void function_9(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_registry *wthp_registry = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *name = (void*)(body + PADDED (sizeof (uint32_t)));
  unsigned int interface_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  const char *interface = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int));
  uint32_t *version = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int) + PADDED (interface_sz) );

  wth_debug ("wthp_registry_send_global(%u, %u, [variable type const char *], %u) (opcode 9) called.", ((struct wth_object *)wthp_registry)->id, *name, *version);
  START_TIMING();
  ((struct wthp_registry_listener *)(((struct wth_object *)wthp_registry)->vfunc))->global
    (wthp_registry, *name, interface, *version);
  END_TIMING("wthp_registry_send_global");
}

/* Function: wthp_registry_send_global_remove
 * Opcode 10 (hexadecimal: 0xa)
 */
static void function_10(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_registry *wthp_registry = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *name = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_registry_send_global_remove(%u, %u) (opcode 10) called.", ((struct wth_object *)wthp_registry)->id, *name);
  START_TIMING();
  ((struct wthp_registry_listener *)(((struct wth_object *)wthp_registry)->vfunc))->global_remove
    (wthp_registry, *name);
  END_TIMING("wthp_registry_send_global_remove");
}

/* Function: wthp_callback_send_done
 * Opcode 11 (hexadecimal: 0xb)
 */
static void function_11(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_callback *wthp_callback = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *callback_data = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_callback_send_done(%u, %u) (opcode 11) called.", ((struct wth_object *)wthp_callback)->id, *callback_data);
  START_TIMING();
  ((struct wthp_callback_listener *)(((struct wth_object *)wthp_callback)->vfunc))->done
    (wthp_callback, *callback_data);
  END_TIMING("wthp_callback_send_done");
}

/* Function: wthp_blob_factory_send_format
 * Opcode 15 (hexadecimal: 0xf)
 */
static void function_15(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_blob_factory *wthp_blob_factory = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *format = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_blob_factory_send_format(%u, %u) (opcode 15) called.", ((struct wth_object *)wthp_blob_factory)->id, *format);
  START_TIMING();
  ((struct wthp_blob_factory_listener *)(((struct wth_object *)wthp_blob_factory)->vfunc))->format
    (wthp_blob_factory, *format);
  END_TIMING("wthp_blob_factory_send_format");
}

/* Function: wthp_buffer_send_complete
 * Opcode 17 (hexadecimal: 0x11)
 */
static void function_17(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_buffer *wthp_buffer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *serial = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_buffer_send_complete(%u, %u) (opcode 17) called.", ((struct wth_object *)wthp_buffer)->id, *serial);
  START_TIMING();
  ((struct wthp_buffer_listener *)(((struct wth_object *)wthp_buffer)->vfunc))->complete
    (wthp_buffer, *serial);
  END_TIMING("wthp_buffer_send_complete");
}

/* Function: wthp_surface_send_enter
 * Opcode 25 (hexadecimal: 0x19)
 */
static void function_25(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_output *output = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_surface_send_enter(%u, %u) (opcode 25) called.", ((struct wth_object *)wthp_surface)->id, ((struct wth_object *)output)->id);
  START_TIMING();
  ((struct wthp_surface_listener *)(((struct wth_object *)wthp_surface)->vfunc))->enter
    (wthp_surface, output);
  END_TIMING("wthp_surface_send_enter");
}

/* Function: wthp_surface_send_leave
 * Opcode 26 (hexadecimal: 0x1a)
 */
static void function_26(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_output *output = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_surface_send_leave(%u, %u) (opcode 26) called.", ((struct wth_object *)wthp_surface)->id, ((struct wth_object *)output)->id);
  START_TIMING();
  ((struct wthp_surface_listener *)(((struct wth_object *)wthp_surface)->vfunc))->leave
    (wthp_surface, output);
  END_TIMING("wthp_surface_send_leave");
}

/* Function: wthp_seat_send_capabilities
 * Opcode 30 (hexadecimal: 0x1e)
 */
static void function_30(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_seat *wthp_seat = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *capabilities = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_seat_send_capabilities(%u, %u) (opcode 30) called.", ((struct wth_object *)wthp_seat)->id, *capabilities);
  START_TIMING();
  ((struct wthp_seat_listener *)(((struct wth_object *)wthp_seat)->vfunc))->capabilities
    (wthp_seat, *capabilities);
  END_TIMING("wthp_seat_send_capabilities");
}

/* Function: wthp_seat_send_name
 * Opcode 34 (hexadecimal: 0x22)
 */
static void function_34(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_seat *wthp_seat = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  unsigned int name_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)));
  const char *name = (void*)(body + PADDED (sizeof (uint32_t)) + sizeof (unsigned int));

  wth_debug ("wthp_seat_send_name(%u, [variable type const char *]) (opcode 34) called.", ((struct wth_object *)wthp_seat)->id);
  START_TIMING();
  ((struct wthp_seat_listener *)(((struct wth_object *)wthp_seat)->vfunc))->name
    (wthp_seat, name);
  END_TIMING("wthp_seat_send_name");
}

/* Function: wthp_pointer_send_enter
 * Opcode 37 (hexadecimal: 0x25)
 */
static void function_37(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_pointer *wthp_pointer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *serial = (void*)(body + PADDED (sizeof (uint32_t)));
  struct wthp_surface *surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t))));
  wth_fixed_t *surface_x = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  wth_fixed_t *surface_y = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (wth_fixed_t)));

  wth_debug ("wthp_pointer_send_enter(%u, %u, %u, [unprintable type wth_fixed_t], [unprintable type wth_fixed_t]) (opcode 37) called.", ((struct wth_object *)wthp_pointer)->id, *serial, ((struct wth_object *)surface)->id);
  START_TIMING();
  ((struct wthp_pointer_listener *)(((struct wth_object *)wthp_pointer)->vfunc))->enter
    (wthp_pointer, *serial, surface, *surface_x, *surface_y);
  END_TIMING("wthp_pointer_send_enter");
}

/* Function: wthp_pointer_send_leave
 * Opcode 38 (hexadecimal: 0x26)
 */
static void function_38(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_pointer *wthp_pointer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *serial = (void*)(body + PADDED (sizeof (uint32_t)));
  struct wthp_surface *surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_pointer_send_leave(%u, %u, %u) (opcode 38) called.", ((struct wth_object *)wthp_pointer)->id, *serial, ((struct wth_object *)surface)->id);
  START_TIMING();
  ((struct wthp_pointer_listener *)(((struct wth_object *)wthp_pointer)->vfunc))->leave
    (wthp_pointer, *serial, surface);
  END_TIMING("wthp_pointer_send_leave");
}

/* Function: wthp_pointer_send_motion
 * Opcode 39 (hexadecimal: 0x27)
 */
static void function_39(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_pointer *wthp_pointer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *time = (void*)(body + PADDED (sizeof (uint32_t)));
  wth_fixed_t *surface_x = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  wth_fixed_t *surface_y = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (wth_fixed_t)));

  wth_debug ("wthp_pointer_send_motion(%u, %u, [unprintable type wth_fixed_t], [unprintable type wth_fixed_t]) (opcode 39) called.", ((struct wth_object *)wthp_pointer)->id, *time);
  START_TIMING();
  ((struct wthp_pointer_listener *)(((struct wth_object *)wthp_pointer)->vfunc))->motion
    (wthp_pointer, *time, *surface_x, *surface_y);
  END_TIMING("wthp_pointer_send_motion");
}

/* Function: wthp_pointer_send_button
 * Opcode 40 (hexadecimal: 0x28)
 */
static void function_40(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_pointer *wthp_pointer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *serial = (void*)(body + PADDED (sizeof (uint32_t)));
  uint32_t *time = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  uint32_t *button = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  uint32_t *state = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_pointer_send_button(%u, %u, %u, %u, %u) (opcode 40) called.", ((struct wth_object *)wthp_pointer)->id, *serial, *time, *button, *state);
  START_TIMING();
  ((struct wthp_pointer_listener *)(((struct wth_object *)wthp_pointer)->vfunc))->button
    (wthp_pointer, *serial, *time, *button, *state);
  END_TIMING("wthp_pointer_send_button");
}

/* Function: wthp_pointer_send_axis
 * Opcode 41 (hexadecimal: 0x29)
 */
static void function_41(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_pointer *wthp_pointer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *time = (void*)(body + PADDED (sizeof (uint32_t)));
  uint32_t *axis = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  wth_fixed_t *value = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_pointer_send_axis(%u, %u, %u, [unprintable type wth_fixed_t]) (opcode 41) called.", ((struct wth_object *)wthp_pointer)->id, *time, *axis);
  START_TIMING();
  ((struct wthp_pointer_listener *)(((struct wth_object *)wthp_pointer)->vfunc))->axis
    (wthp_pointer, *time, *axis, *value);
  END_TIMING("wthp_pointer_send_axis");
}

/* Function: wthp_pointer_send_frame
 * Opcode 43 (hexadecimal: 0x2b)
 */
static void function_43(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_pointer *wthp_pointer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_pointer_send_frame(%u) (opcode 43) called.", ((struct wth_object *)wthp_pointer)->id);
  START_TIMING();
  ((struct wthp_pointer_listener *)(((struct wth_object *)wthp_pointer)->vfunc))->frame
    (wthp_pointer);
  END_TIMING("wthp_pointer_send_frame");
}

/* Function: wthp_pointer_send_axis_source
 * Opcode 44 (hexadecimal: 0x2c)
 */
static void function_44(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_pointer *wthp_pointer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *axis_source = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_pointer_send_axis_source(%u, %u) (opcode 44) called.", ((struct wth_object *)wthp_pointer)->id, *axis_source);
  START_TIMING();
  ((struct wthp_pointer_listener *)(((struct wth_object *)wthp_pointer)->vfunc))->axis_source
    (wthp_pointer, *axis_source);
  END_TIMING("wthp_pointer_send_axis_source");
}

/* Function: wthp_pointer_send_axis_stop
 * Opcode 45 (hexadecimal: 0x2d)
 */
static void function_45(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_pointer *wthp_pointer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *time = (void*)(body + PADDED (sizeof (uint32_t)));
  uint32_t *axis = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_pointer_send_axis_stop(%u, %u, %u) (opcode 45) called.", ((struct wth_object *)wthp_pointer)->id, *time, *axis);
  START_TIMING();
  ((struct wthp_pointer_listener *)(((struct wth_object *)wthp_pointer)->vfunc))->axis_stop
    (wthp_pointer, *time, *axis);
  END_TIMING("wthp_pointer_send_axis_stop");
}

/* Function: wthp_pointer_send_axis_discrete
 * Opcode 46 (hexadecimal: 0x2e)
 */
static void function_46(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_pointer *wthp_pointer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *axis = (void*)(body + PADDED (sizeof (uint32_t)));
  int32_t *discrete = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_pointer_send_axis_discrete(%u, %u, %d) (opcode 46) called.", ((struct wth_object *)wthp_pointer)->id, *axis, *discrete);
  START_TIMING();
  ((struct wthp_pointer_listener *)(((struct wth_object *)wthp_pointer)->vfunc))->axis_discrete
    (wthp_pointer, *axis, *discrete);
  END_TIMING("wthp_pointer_send_axis_discrete");
}

/* Function: wthp_keyboard_send_keymap
 * Opcode 47 (hexadecimal: 0x2f)
 */
static void function_47(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_keyboard *wthp_keyboard = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *format = (void*)(body + PADDED (sizeof (uint32_t)));
  unsigned int keymap_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  void *keymap = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int));

  wth_debug ("wthp_keyboard_send_keymap(%u, %u, [variable type uint32_t], [variable type void *]) (opcode 47) called.", ((struct wth_object *)wthp_keyboard)->id, *format);
  START_TIMING();
  ((struct wthp_keyboard_listener *)(((struct wth_object *)wthp_keyboard)->vfunc))->keymap
    (wthp_keyboard, *format, keymap_sz, keymap);
  END_TIMING("wthp_keyboard_send_keymap");
}

/* Function: wthp_keyboard_send_enter
 * Opcode 48 (hexadecimal: 0x30)
 */
static void function_48(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_keyboard *wthp_keyboard = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *serial = (void*)(body + PADDED (sizeof (uint32_t)));
  struct wthp_surface *surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t))));
  unsigned int keys_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  struct wth_array keys = { keys_sz, keys_sz, (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int)) };

  wth_debug ("wthp_keyboard_send_enter(%u, %u, %u, [variable type struct wth_array *]) (opcode 48) called.", ((struct wth_object *)wthp_keyboard)->id, *serial, ((struct wth_object *)surface)->id);
  START_TIMING();
  ((struct wthp_keyboard_listener *)(((struct wth_object *)wthp_keyboard)->vfunc))->enter
    (wthp_keyboard, *serial, surface, &keys);
  END_TIMING("wthp_keyboard_send_enter");
}

/* Function: wthp_keyboard_send_leave
 * Opcode 49 (hexadecimal: 0x31)
 */
static void function_49(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_keyboard *wthp_keyboard = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *serial = (void*)(body + PADDED (sizeof (uint32_t)));
  struct wthp_surface *surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_keyboard_send_leave(%u, %u, %u) (opcode 49) called.", ((struct wth_object *)wthp_keyboard)->id, *serial, ((struct wth_object *)surface)->id);
  START_TIMING();
  ((struct wthp_keyboard_listener *)(((struct wth_object *)wthp_keyboard)->vfunc))->leave
    (wthp_keyboard, *serial, surface);
  END_TIMING("wthp_keyboard_send_leave");
}

/* Function: wthp_keyboard_send_key
 * Opcode 50 (hexadecimal: 0x32)
 */
static void function_50(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_keyboard *wthp_keyboard = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *serial = (void*)(body + PADDED (sizeof (uint32_t)));
  uint32_t *time = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  uint32_t *key = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  uint32_t *state = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_keyboard_send_key(%u, %u, %u, %u, %u) (opcode 50) called.", ((struct wth_object *)wthp_keyboard)->id, *serial, *time, *key, *state);
  START_TIMING();
  ((struct wthp_keyboard_listener *)(((struct wth_object *)wthp_keyboard)->vfunc))->key
    (wthp_keyboard, *serial, *time, *key, *state);
  END_TIMING("wthp_keyboard_send_key");
}

/* Function: wthp_keyboard_send_modifiers
 * Opcode 51 (hexadecimal: 0x33)
 */
static void function_51(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_keyboard *wthp_keyboard = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *serial = (void*)(body + PADDED (sizeof (uint32_t)));
  uint32_t *mods_depressed = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  uint32_t *mods_latched = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  uint32_t *mods_locked = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  uint32_t *group = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_keyboard_send_modifiers(%u, %u, %u, %u, %u, %u) (opcode 51) called.", ((struct wth_object *)wthp_keyboard)->id, *serial, *mods_depressed, *mods_latched, *mods_locked, *group);
  START_TIMING();
  ((struct wthp_keyboard_listener *)(((struct wth_object *)wthp_keyboard)->vfunc))->modifiers
    (wthp_keyboard, *serial, *mods_depressed, *mods_latched, *mods_locked, *group);
  END_TIMING("wthp_keyboard_send_modifiers");
}

/* Function: wthp_keyboard_send_repeat_info
 * Opcode 53 (hexadecimal: 0x35)
 */
static void function_53(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_keyboard *wthp_keyboard = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  int32_t *rate = (void*)(body + PADDED (sizeof (uint32_t)));
  int32_t *delay = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)));

  wth_debug ("wthp_keyboard_send_repeat_info(%u, %d, %d) (opcode 53) called.", ((struct wth_object *)wthp_keyboard)->id, *rate, *delay);
  START_TIMING();
  ((struct wthp_keyboard_listener *)(((struct wth_object *)wthp_keyboard)->vfunc))->repeat_info
    (wthp_keyboard, *rate, *delay);
  END_TIMING("wthp_keyboard_send_repeat_info");
}

/* Function: wthp_touch_send_down
 * Opcode 54 (hexadecimal: 0x36)
 */
static void function_54(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_touch *wthp_touch = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *serial = (void*)(body + PADDED (sizeof (uint32_t)));
  uint32_t *time = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  struct wthp_surface *surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t))));
  int32_t *id = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  wth_fixed_t *x = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)));
  wth_fixed_t *y = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (wth_fixed_t)));

  wth_debug ("wthp_touch_send_down(%u, %u, %u, %u, %d, [unprintable type wth_fixed_t], [unprintable type wth_fixed_t]) (opcode 54) called.", ((struct wth_object *)wthp_touch)->id, *serial, *time, ((struct wth_object *)surface)->id, *id);
  START_TIMING();
  ((struct wthp_touch_listener *)(((struct wth_object *)wthp_touch)->vfunc))->down
    (wthp_touch, *serial, *time, surface, *id, *x, *y);
  END_TIMING("wthp_touch_send_down");
}

/* Function: wthp_touch_send_up
 * Opcode 55 (hexadecimal: 0x37)
 */
static void function_55(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_touch *wthp_touch = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *serial = (void*)(body + PADDED (sizeof (uint32_t)));
  uint32_t *time = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  int32_t *id = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_touch_send_up(%u, %u, %u, %d) (opcode 55) called.", ((struct wth_object *)wthp_touch)->id, *serial, *time, *id);
  START_TIMING();
  ((struct wthp_touch_listener *)(((struct wth_object *)wthp_touch)->vfunc))->up
    (wthp_touch, *serial, *time, *id);
  END_TIMING("wthp_touch_send_up");
}

/* Function: wthp_touch_send_motion
 * Opcode 56 (hexadecimal: 0x38)
 */
static void function_56(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_touch *wthp_touch = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *time = (void*)(body + PADDED (sizeof (uint32_t)));
  int32_t *id = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  wth_fixed_t *x = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)));
  wth_fixed_t *y = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (wth_fixed_t)));

  wth_debug ("wthp_touch_send_motion(%u, %u, %d, [unprintable type wth_fixed_t], [unprintable type wth_fixed_t]) (opcode 56) called.", ((struct wth_object *)wthp_touch)->id, *time, *id);
  START_TIMING();
  ((struct wthp_touch_listener *)(((struct wth_object *)wthp_touch)->vfunc))->motion
    (wthp_touch, *time, *id, *x, *y);
  END_TIMING("wthp_touch_send_motion");
}

/* Function: wthp_touch_send_frame
 * Opcode 57 (hexadecimal: 0x39)
 */
static void function_57(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_touch *wthp_touch = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_touch_send_frame(%u) (opcode 57) called.", ((struct wth_object *)wthp_touch)->id);
  START_TIMING();
  ((struct wthp_touch_listener *)(((struct wth_object *)wthp_touch)->vfunc))->frame
    (wthp_touch);
  END_TIMING("wthp_touch_send_frame");
}

/* Function: wthp_touch_send_cancel
 * Opcode 58 (hexadecimal: 0x3a)
 */
static void function_58(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_touch *wthp_touch = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_touch_send_cancel(%u) (opcode 58) called.", ((struct wth_object *)wthp_touch)->id);
  START_TIMING();
  ((struct wthp_touch_listener *)(((struct wth_object *)wthp_touch)->vfunc))->cancel
    (wthp_touch);
  END_TIMING("wthp_touch_send_cancel");
}

/* Function: wthp_output_send_geometry
 * Opcode 60 (hexadecimal: 0x3c)
 */
static void function_60(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_output *wthp_output = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  int32_t *x = (void*)(body + PADDED (sizeof (uint32_t)));
  int32_t *y = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)));
  int32_t *physical_width = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));
  int32_t *physical_height = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));
  int32_t *subpixel = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));
  unsigned int make_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));
  const char *make = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + sizeof (unsigned int));
  unsigned int model_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + sizeof (unsigned int) + PADDED (make_sz) );
  const char *model = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + sizeof (unsigned int) + PADDED (make_sz)  + sizeof (unsigned int));
  int32_t *transform = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + sizeof (unsigned int) + PADDED (make_sz)  + sizeof (unsigned int) + PADDED (model_sz) );

  wth_debug ("wthp_output_send_geometry(%u, %d, %d, %d, %d, %d, [variable type const char *], [variable type const char *], %d) (opcode 60) called.", ((struct wth_object *)wthp_output)->id, *x, *y, *physical_width, *physical_height, *subpixel, *transform);
  START_TIMING();
  ((struct wthp_output_listener *)(((struct wth_object *)wthp_output)->vfunc))->geometry
    (wthp_output, *x, *y, *physical_width, *physical_height, *subpixel, make, model, *transform);
  END_TIMING("wthp_output_send_geometry");
}

/* Function: wthp_output_send_mode
 * Opcode 61 (hexadecimal: 0x3d)
 */
static void function_61(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_output *wthp_output = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *flags = (void*)(body + PADDED (sizeof (uint32_t)));
  int32_t *width = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  int32_t *height = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)));
  int32_t *refresh = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));

  wth_debug ("wthp_output_send_mode(%u, %u, %d, %d, %d) (opcode 61) called.", ((struct wth_object *)wthp_output)->id, *flags, *width, *height, *refresh);
  START_TIMING();
  ((struct wthp_output_listener *)(((struct wth_object *)wthp_output)->vfunc))->mode
    (wthp_output, *flags, *width, *height, *refresh);
  END_TIMING("wthp_output_send_mode");
}

/* Function: wthp_output_send_done
 * Opcode 62 (hexadecimal: 0x3e)
 */
static void function_62(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_output *wthp_output = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_output_send_done(%u) (opcode 62) called.", ((struct wth_object *)wthp_output)->id);
  START_TIMING();
  ((struct wthp_output_listener *)(((struct wth_object *)wthp_output)->vfunc))->done
    (wthp_output);
  END_TIMING("wthp_output_send_done");
}

/* Function: wthp_output_send_scale
 * Opcode 63 (hexadecimal: 0x3f)
 */
static void function_63(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_output *wthp_output = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  int32_t *factor = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_output_send_scale(%u, %d) (opcode 63) called.", ((struct wth_object *)wthp_output)->id, *factor);
  START_TIMING();
  ((struct wthp_output_listener *)(((struct wth_object *)wthp_output)->vfunc))->scale
    (wthp_output, *factor);
  END_TIMING("wthp_output_send_scale");
}

/* Function: wthp_farstream_remote_send_port
 * Opcode 72 (hexadecimal: 0x48)
 */
static void function_72(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_farstream_remote *wthp_farstream_remote = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *port = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_farstream_remote_send_port(%u, %u) (opcode 72) called.", ((struct wth_object *)wthp_farstream_remote)->id, *port);
  START_TIMING();
  ((struct wthp_farstream_remote_listener *)(((struct wth_object *)wthp_farstream_remote)->vfunc))->port
    (wthp_farstream_remote, *port);
  END_TIMING("wthp_farstream_remote_send_port");
}

/* Function: wthp_farstream_remote_send_codec_answer
 * Opcode 73 (hexadecimal: 0x49)
 */
static void function_73(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_farstream_remote *wthp_farstream_remote = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  unsigned int list_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)));
  const char *list = (void*)(body + PADDED (sizeof (uint32_t)) + sizeof (unsigned int));

  wth_debug ("wthp_farstream_remote_send_codec_answer(%u, [variable type const char *]) (opcode 73) called.", ((struct wth_object *)wthp_farstream_remote)->id);
  START_TIMING();
  ((struct wthp_farstream_remote_listener *)(((struct wth_object *)wthp_farstream_remote)->vfunc))->codec_answer
    (wthp_farstream_remote, list);
  END_TIMING("wthp_farstream_remote_send_codec_answer");
}

/* Function: wthp_farstream_remote_send_error
 * Opcode 74 (hexadecimal: 0x4a)
 */
static void function_74(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_farstream_remote *wthp_farstream_remote = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *code = (void*)(body + PADDED (sizeof (uint32_t)));
  unsigned int message_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  const char *message = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int));

  wth_debug ("wthp_farstream_remote_send_error(%u, %u, [variable type const char *]) (opcode 74) called.", ((struct wth_object *)wthp_farstream_remote)->id, *code);
  START_TIMING();
  ((struct wthp_farstream_remote_listener *)(((struct wth_object *)wthp_farstream_remote)->vfunc))->error
    (wthp_farstream_remote, *code, message);
  END_TIMING("wthp_farstream_remote_send_error");
}

const int demarshaller_max_opcode = 74;
const demarshaller_helper_function_t event_demarshaller_functions[] = {
  NULL,
  NULL,
  NULL,
  NULL,
  function_4,
  function_5,
  function_6,
  NULL,
  NULL,
  function_9,
  function_10,
  function_11,
  NULL,
  NULL,
  NULL,
  function_15,
  NULL,
  function_17,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  function_25,
  function_26,
  NULL,
  NULL,
  NULL,
  function_30,
  NULL,
  NULL,
  NULL,
  function_34,
  NULL,
  NULL,
  function_37,
  function_38,
  function_39,
  function_40,
  function_41,
  NULL,
  function_43,
  function_44,
  function_45,
  function_46,
  function_47,
  function_48,
  function_49,
  function_50,
  function_51,
  NULL,
  function_53,
  function_54,
  function_55,
  function_56,
  function_57,
  function_58,
  NULL,
  function_60,
  function_61,
  function_62,
  function_63,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  function_72,
  function_73,
  function_74,
};
