/*
 * Copyright © 2013-2014 Collabora, Ltd.
 * Copyright © 2016 DENSO CORPORATION
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "waltham-object.h"
#include "waltham-connection.h"
#include "waltham-private.h"

#include "marshaller_log.h"
#include "marshaller.h"

#include "waltham-client.h"

WTH_EXPORT void
wth_display_client_version (struct wth_display * wth_display, uint32_t client_version)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wth_display_client_version", "[return/variable/unprintable type struct wth_display *], %u", client_version);
   START_MESSAGE("wth_display_client_version", sz, 1);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wth_display)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&client_version, sizeof(client_version) );
   END_MESSAGE(((struct wth_object *)wth_display)->connection);
   END_TIMING("");
}

WTH_EXPORT struct wthp_callback *
wth_display_sync (struct wth_display * wth_display)
{
   struct wthp_callback * ret = (struct wthp_callback *) wth_object_new (((struct wth_object *)wth_display)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wth_display_sync", "[return/variable/unprintable type struct wth_display *]");
   START_MESSAGE("wth_display_sync", sz, 2);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wth_display)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wth_display)->connection);
   END_TIMING("");
   return ret;
}

WTH_EXPORT struct wthp_registry *
wth_display_get_registry (struct wth_display * wth_display)
{
   struct wthp_registry * ret = (struct wthp_registry *) wth_object_new (((struct wth_object *)wth_display)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wth_display_get_registry", "[return/variable/unprintable type struct wth_display *]");
   START_MESSAGE("wth_display_get_registry", sz, 3);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wth_display)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wth_display)->connection);
   END_TIMING("");
   return ret;
}

WTH_EXPORT void
wthp_registry_destroy (struct wthp_registry * wthp_registry)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_registry_destroy", "[return/variable/unprintable type struct wthp_registry *]");
   START_MESSAGE("wthp_registry_destroy", sz, 7);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_registry)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_registry)->connection);
   wthp_registry_free(wthp_registry);
   END_TIMING("");
}

WTH_EXPORT struct wth_object *
wthp_registry_bind (struct wthp_registry * wthp_registry, uint32_t name, const char * interface, uint32_t version)
{
   struct wth_object * ret = (struct wth_object *) wth_object_new (((struct wth_object *)wthp_registry)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   int interface_sz = strlen (interface) + 1;
   sz += sizeof (unsigned int) + PADDED(interface_sz);

   START_TIMING("wthp_registry_bind", "[return/variable/unprintable type struct wthp_registry *], %u, %s, %u", name, interface, version);
   START_MESSAGE("wthp_registry_bind", sz, 8);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_registry)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&name, sizeof(name) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   SERIALIZE_DATA( (void *)interface, interface_sz);
   SERIALIZE_PARAM( (void *)&version, sizeof(version) );
   END_MESSAGE(((struct wth_object *)wthp_registry)->connection);
   END_TIMING("");
   return ret;
}

WTH_EXPORT struct wthp_surface *
wthp_compositor_create_surface (struct wthp_compositor * wthp_compositor)
{
   struct wthp_surface * ret = (struct wthp_surface *) wth_object_new (((struct wth_object *)wthp_compositor)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_compositor_create_surface", "[return/variable/unprintable type struct wthp_compositor *]");
   START_MESSAGE("wthp_compositor_create_surface", sz, 12);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_compositor)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_compositor)->connection);
   END_TIMING("");
   return ret;
}

WTH_EXPORT struct wthp_region *
wthp_compositor_create_region (struct wthp_compositor * wthp_compositor)
{
   struct wthp_region * ret = (struct wthp_region *) wth_object_new (((struct wth_object *)wthp_compositor)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_compositor_create_region", "[return/variable/unprintable type struct wthp_compositor *]");
   START_MESSAGE("wthp_compositor_create_region", sz, 13);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_compositor)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_compositor)->connection);
   END_TIMING("");
   return ret;
}

WTH_EXPORT struct wthp_buffer *
wthp_blob_factory_create_buffer (struct wthp_blob_factory * wthp_blob_factory, uint32_t data_sz, void * data, int32_t width, int32_t height, int32_t stride, uint32_t format)
{
   struct wthp_buffer * ret = (struct wthp_buffer *) wth_object_new (((struct wth_object *)wthp_blob_factory)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(data_sz) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_blob_factory_create_buffer", "[return/variable/unprintable type struct wthp_blob_factory *], %u, [return/variable/unprintable type void *], %d, %d, %d, %u", data_sz, width, height, stride, format);
   START_MESSAGE("wthp_blob_factory_create_buffer", sz, 14);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_blob_factory)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   SERIALIZE_DATA( (void *)data, data_sz);
   SERIALIZE_PARAM( (void *)&width, sizeof(width) );
   SERIALIZE_PARAM( (void *)&height, sizeof(height) );
   SERIALIZE_PARAM( (void *)&stride, sizeof(stride) );
   SERIALIZE_PARAM( (void *)&format, sizeof(format) );
   END_MESSAGE(((struct wth_object *)wthp_blob_factory)->connection);
   END_TIMING("");
   return ret;
}

WTH_EXPORT void
wthp_buffer_destroy (struct wthp_buffer * wthp_buffer)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_buffer_destroy", "[return/variable/unprintable type struct wthp_buffer *]");
   START_MESSAGE("wthp_buffer_destroy", sz, 16);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_buffer)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_buffer)->connection);
   wthp_buffer_free(wthp_buffer);
   END_TIMING("");
}

WTH_EXPORT void
wthp_surface_destroy (struct wthp_surface * wthp_surface)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_surface_destroy", "[return/variable/unprintable type struct wthp_surface *]");
   START_MESSAGE("wthp_surface_destroy", sz, 18);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   wthp_surface_free(wthp_surface);
   END_TIMING("");
}

WTH_EXPORT void
wthp_surface_attach (struct wthp_surface * wthp_surface, struct wthp_buffer * buffer, int32_t x, int32_t y)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_surface_attach", "[return/variable/unprintable type struct wthp_surface *], [return/variable/unprintable type struct wthp_buffer *], %d, %d", x, y);
   START_MESSAGE("wthp_surface_attach", sz, 19);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)buffer)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&x, sizeof(x) );
   SERIALIZE_PARAM( (void *)&y, sizeof(y) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_surface_damage (struct wthp_surface * wthp_surface, int32_t x, int32_t y, int32_t width, int32_t height)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_surface_damage", "[return/variable/unprintable type struct wthp_surface *], %d, %d, %d, %d", x, y, width, height);
   START_MESSAGE("wthp_surface_damage", sz, 20);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&x, sizeof(x) );
   SERIALIZE_PARAM( (void *)&y, sizeof(y) );
   SERIALIZE_PARAM( (void *)&width, sizeof(width) );
   SERIALIZE_PARAM( (void *)&height, sizeof(height) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   END_TIMING("");
}

WTH_EXPORT struct wthp_callback *
wthp_surface_frame (struct wthp_surface * wthp_surface)
{
   struct wthp_callback * ret = (struct wthp_callback *) wth_object_new (((struct wth_object *)wthp_surface)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_surface_frame", "[return/variable/unprintable type struct wthp_surface *]");
   START_MESSAGE("wthp_surface_frame", sz, 21);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   END_TIMING("");
   return ret;
}

WTH_EXPORT void
wthp_surface_set_opaque_region (struct wthp_surface * wthp_surface, struct wthp_region * region)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_surface_set_opaque_region", "[return/variable/unprintable type struct wthp_surface *], [return/variable/unprintable type struct wthp_region *]");
   START_MESSAGE("wthp_surface_set_opaque_region", sz, 22);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)region)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_surface_set_input_region (struct wthp_surface * wthp_surface, struct wthp_region * region)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_surface_set_input_region", "[return/variable/unprintable type struct wthp_surface *], [return/variable/unprintable type struct wthp_region *]");
   START_MESSAGE("wthp_surface_set_input_region", sz, 23);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)region)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_surface_commit (struct wthp_surface * wthp_surface)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_surface_commit", "[return/variable/unprintable type struct wthp_surface *]");
   START_MESSAGE("wthp_surface_commit", sz, 24);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_surface_set_buffer_transform (struct wthp_surface * wthp_surface, int32_t transform)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_surface_set_buffer_transform", "[return/variable/unprintable type struct wthp_surface *], %d", transform);
   START_MESSAGE("wthp_surface_set_buffer_transform", sz, 27);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&transform, sizeof(transform) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_surface_set_buffer_scale (struct wthp_surface * wthp_surface, int32_t scale)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_surface_set_buffer_scale", "[return/variable/unprintable type struct wthp_surface *], %d", scale);
   START_MESSAGE("wthp_surface_set_buffer_scale", sz, 28);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&scale, sizeof(scale) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_surface_damage_buffer (struct wthp_surface * wthp_surface, int32_t x, int32_t y, int32_t width, int32_t height)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_surface_damage_buffer", "[return/variable/unprintable type struct wthp_surface *], %d, %d, %d, %d", x, y, width, height);
   START_MESSAGE("wthp_surface_damage_buffer", sz, 29);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&x, sizeof(x) );
   SERIALIZE_PARAM( (void *)&y, sizeof(y) );
   SERIALIZE_PARAM( (void *)&width, sizeof(width) );
   SERIALIZE_PARAM( (void *)&height, sizeof(height) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   END_TIMING("");
}

WTH_EXPORT struct wthp_pointer *
wthp_seat_get_pointer (struct wthp_seat * wthp_seat)
{
   struct wthp_pointer * ret = (struct wthp_pointer *) wth_object_new (((struct wth_object *)wthp_seat)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_seat_get_pointer", "[return/variable/unprintable type struct wthp_seat *]");
   START_MESSAGE("wthp_seat_get_pointer", sz, 31);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_seat)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_seat)->connection);
   END_TIMING("");
   return ret;
}

WTH_EXPORT struct wthp_keyboard *
wthp_seat_get_keyboard (struct wthp_seat * wthp_seat)
{
   struct wthp_keyboard * ret = (struct wthp_keyboard *) wth_object_new (((struct wth_object *)wthp_seat)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_seat_get_keyboard", "[return/variable/unprintable type struct wthp_seat *]");
   START_MESSAGE("wthp_seat_get_keyboard", sz, 32);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_seat)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_seat)->connection);
   END_TIMING("");
   return ret;
}

WTH_EXPORT struct wthp_touch *
wthp_seat_get_touch (struct wthp_seat * wthp_seat)
{
   struct wthp_touch * ret = (struct wthp_touch *) wth_object_new (((struct wth_object *)wthp_seat)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_seat_get_touch", "[return/variable/unprintable type struct wthp_seat *]");
   START_MESSAGE("wthp_seat_get_touch", sz, 33);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_seat)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_seat)->connection);
   END_TIMING("");
   return ret;
}

WTH_EXPORT void
wthp_seat_release (struct wthp_seat * wthp_seat)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_seat_release", "[return/variable/unprintable type struct wthp_seat *]");
   START_MESSAGE("wthp_seat_release", sz, 35);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_seat)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_seat)->connection);
   wthp_seat_free(wthp_seat);
   END_TIMING("");
}

WTH_EXPORT void
wthp_pointer_set_cursor (struct wthp_pointer * wthp_pointer, uint32_t serial, struct wthp_surface * surface, int32_t hotspot_x, int32_t hotspot_y)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_pointer_set_cursor", "[return/variable/unprintable type struct wthp_pointer *], %u, [return/variable/unprintable type struct wthp_surface *], %d, %d", serial, hotspot_x, hotspot_y);
   START_MESSAGE("wthp_pointer_set_cursor", sz, 36);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_pointer)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&serial, sizeof(serial) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&hotspot_x, sizeof(hotspot_x) );
   SERIALIZE_PARAM( (void *)&hotspot_y, sizeof(hotspot_y) );
   END_MESSAGE(((struct wth_object *)wthp_pointer)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_pointer_release (struct wthp_pointer * wthp_pointer)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_pointer_release", "[return/variable/unprintable type struct wthp_pointer *]");
   START_MESSAGE("wthp_pointer_release", sz, 42);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_pointer)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_pointer)->connection);
   wthp_pointer_free(wthp_pointer);
   END_TIMING("");
}

WTH_EXPORT void
wthp_keyboard_release (struct wthp_keyboard * wthp_keyboard)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_keyboard_release", "[return/variable/unprintable type struct wthp_keyboard *]");
   START_MESSAGE("wthp_keyboard_release", sz, 52);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_keyboard)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_keyboard)->connection);
   wthp_keyboard_free(wthp_keyboard);
   END_TIMING("");
}

WTH_EXPORT void
wthp_touch_release (struct wthp_touch * wthp_touch)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_touch_release", "[return/variable/unprintable type struct wthp_touch *]");
   START_MESSAGE("wthp_touch_release", sz, 59);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_touch)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_touch)->connection);
   wthp_touch_free(wthp_touch);
   END_TIMING("");
}

WTH_EXPORT void
wthp_region_destroy (struct wthp_region * wthp_region)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_region_destroy", "[return/variable/unprintable type struct wthp_region *]");
   START_MESSAGE("wthp_region_destroy", sz, 64);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_region)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_region)->connection);
   wthp_region_free(wthp_region);
   END_TIMING("");
}

WTH_EXPORT void
wthp_region_add (struct wthp_region * wthp_region, int32_t x, int32_t y, int32_t width, int32_t height)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_region_add", "[return/variable/unprintable type struct wthp_region *], %d, %d, %d, %d", x, y, width, height);
   START_MESSAGE("wthp_region_add", sz, 65);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_region)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&x, sizeof(x) );
   SERIALIZE_PARAM( (void *)&y, sizeof(y) );
   SERIALIZE_PARAM( (void *)&width, sizeof(width) );
   SERIALIZE_PARAM( (void *)&height, sizeof(height) );
   END_MESSAGE(((struct wth_object *)wthp_region)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_region_subtract (struct wthp_region * wthp_region, int32_t x, int32_t y, int32_t width, int32_t height)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_region_subtract", "[return/variable/unprintable type struct wthp_region *], %d, %d, %d, %d", x, y, width, height);
   START_MESSAGE("wthp_region_subtract", sz, 66);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_region)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&x, sizeof(x) );
   SERIALIZE_PARAM( (void *)&y, sizeof(y) );
   SERIALIZE_PARAM( (void *)&width, sizeof(width) );
   SERIALIZE_PARAM( (void *)&height, sizeof(height) );
   END_MESSAGE(((struct wth_object *)wthp_region)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_farstream_destroy (struct wthp_farstream * wthp_farstream)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_farstream_destroy", "[return/variable/unprintable type struct wthp_farstream *]");
   START_MESSAGE("wthp_farstream_destroy", sz, 67);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_farstream)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_farstream)->connection);
   wthp_farstream_free(wthp_farstream);
   END_TIMING("");
}

WTH_EXPORT struct wthp_farstream_remote *
wthp_farstream_connect (struct wthp_farstream * wthp_farstream, uint32_t port)
{
   struct wthp_farstream_remote * ret = (struct wthp_farstream_remote *) wth_object_new (((struct wth_object *)wthp_farstream)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_farstream_connect", "[return/variable/unprintable type struct wthp_farstream *], %u", port);
   START_MESSAGE("wthp_farstream_connect", sz, 68);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_farstream)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&port, sizeof(port) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_farstream)->connection);
   END_TIMING("");
   return ret;
}

WTH_EXPORT void
wthp_farstream_remote_destroy (struct wthp_farstream_remote * wthp_farstream_remote)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_farstream_remote_destroy", "[return/variable/unprintable type struct wthp_farstream_remote *]");
   START_MESSAGE("wthp_farstream_remote_destroy", sz, 69);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_farstream_remote)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_farstream_remote)->connection);
   wthp_farstream_remote_free(wthp_farstream_remote);
   END_TIMING("");
}

WTH_EXPORT void
wthp_farstream_remote_codec_offer (struct wthp_farstream_remote * wthp_farstream_remote, const char * list)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   int list_sz = strlen (list) + 1;
   sz += sizeof (unsigned int) + PADDED(list_sz);

   START_TIMING("wthp_farstream_remote_codec_offer", "[return/variable/unprintable type struct wthp_farstream_remote *], %s", list);
   START_MESSAGE("wthp_farstream_remote_codec_offer", sz, 70);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_farstream_remote)->id, sizeof(uint32_t) );
   SERIALIZE_DATA( (void *)list, list_sz);
   END_MESSAGE(((struct wth_object *)wthp_farstream_remote)->connection);
   END_TIMING("");
}

WTH_EXPORT struct wthp_buffer *
wthp_farstream_remote_create_buffer (struct wthp_farstream_remote * wthp_farstream_remote, uint32_t frame_id)
{
   struct wthp_buffer * ret = (struct wthp_buffer *) wth_object_new (((struct wth_object *)wthp_farstream_remote)->connection);
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_farstream_remote_create_buffer", "[return/variable/unprintable type struct wthp_farstream_remote *], %u", frame_id);
   START_MESSAGE("wthp_farstream_remote_create_buffer", sz, 71);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_farstream_remote)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&frame_id, sizeof(frame_id) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)ret)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_farstream_remote)->connection);
   END_TIMING("");
   return ret;
}

