/*
 * Copyright © 2013-2014 Collabora, Ltd.
 * Copyright © 2016 DENSO CORPORATION
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "waltham-object.h"
#include "waltham-connection.h"
#include "waltham-private.h"

#include "message.h"
#include "demarshaller.h"

//#define ENABLE_PROFILING 1

// FIXME: we no longer use glib, so this is broken
#ifdef ENABLE_PROFILING
#define START_TIMING() \
  gdouble msec; \
  GTimer *timer = g_timer_new ();
#define END_TIMING(label) \
  msec = g_timer_elapsed (timer, NULL) * 1000; \
  g_timer_destroy (timer); \
  g_debug ("%s (%f ms)", label, msec);
#else
#define START_TIMING()
#define END_TIMING(a)
#endif

#define PADDED(sz) \
   (((sz) + 3) & ~3)

#include "waltham-server.h"

/* Function: wth_display_client_version
 * Opcode 1 (hexadecimal: 0x1)
 */
static void function_1(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wth_display *wth_display = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *client_version = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wth_display_client_version(%u, %u) (opcode 1) called.", ((struct wth_object *)wth_display)->id, *client_version);
  START_TIMING();
  ((struct wth_display_interface *)(((struct wth_object *)wth_display)->vfunc))->client_version
    (wth_display, *client_version);
  END_TIMING("wth_display_client_version");
}

/* Function: wth_display_sync
 * Opcode 2 (hexadecimal: 0x2)
 */
static void function_2(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wth_display *wth_display = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_callback *callback = (struct wthp_callback *) wth_object_new_with_id (((struct wth_object *)wth_display)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wth_display_sync(%u, %u) (opcode 2) called.", ((struct wth_object *)wth_display)->id, ((struct wth_object *)callback)->id);
  START_TIMING();
  ((struct wth_display_interface *)(((struct wth_object *)wth_display)->vfunc))->sync
    (wth_display, callback);
  END_TIMING("wth_display_sync");
}

/* Function: wth_display_get_registry
 * Opcode 3 (hexadecimal: 0x3)
 */
static void function_3(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wth_display *wth_display = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_registry *registry = (struct wthp_registry *) wth_object_new_with_id (((struct wth_object *)wth_display)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wth_display_get_registry(%u, %u) (opcode 3) called.", ((struct wth_object *)wth_display)->id, ((struct wth_object *)registry)->id);
  START_TIMING();
  ((struct wth_display_interface *)(((struct wth_object *)wth_display)->vfunc))->get_registry
    (wth_display, registry);
  END_TIMING("wth_display_get_registry");
}

/* Function: wthp_registry_destroy
 * Opcode 7 (hexadecimal: 0x7)
 */
static void function_7(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_registry *wthp_registry = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_registry_destroy(%u) (opcode 7) called.", ((struct wth_object *)wthp_registry)->id);
  START_TIMING();
  ((struct wthp_registry_interface *)(((struct wth_object *)wthp_registry)->vfunc))->destroy
    (wthp_registry);
  END_TIMING("wthp_registry_destroy");
}

/* Function: wthp_registry_bind
 * Opcode 8 (hexadecimal: 0x8)
 */
static void function_8(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_registry *wthp_registry = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *name = (void*)(body + PADDED (sizeof (uint32_t)));
  struct wth_object *id = (struct wth_object *) wth_object_new_with_id (((struct wth_object *)wthp_registry)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t))));
  unsigned int interface_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  const char *interface = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int));
  uint32_t *version = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int) + PADDED (interface_sz) );

  wth_debug ("wthp_registry_bind(%u, %u, %u, [variable type const char *], %u) (opcode 8) called.", ((struct wth_object *)wthp_registry)->id, *name, ((struct wth_object *)id)->id, *version);
  START_TIMING();
  ((struct wthp_registry_interface *)(((struct wth_object *)wthp_registry)->vfunc))->bind
    (wthp_registry, *name, id, interface, *version);
  END_TIMING("wthp_registry_bind");
}

/* Function: wthp_compositor_create_surface
 * Opcode 12 (hexadecimal: 0xc)
 */
static void function_12(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_compositor *wthp_compositor = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_surface *id = (struct wthp_surface *) wth_object_new_with_id (((struct wth_object *)wthp_compositor)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_compositor_create_surface(%u, %u) (opcode 12) called.", ((struct wth_object *)wthp_compositor)->id, ((struct wth_object *)id)->id);
  START_TIMING();
  ((struct wthp_compositor_interface *)(((struct wth_object *)wthp_compositor)->vfunc))->create_surface
    (wthp_compositor, id);
  END_TIMING("wthp_compositor_create_surface");
}

/* Function: wthp_compositor_create_region
 * Opcode 13 (hexadecimal: 0xd)
 */
static void function_13(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_compositor *wthp_compositor = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_region *id = (struct wthp_region *) wth_object_new_with_id (((struct wth_object *)wthp_compositor)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_compositor_create_region(%u, %u) (opcode 13) called.", ((struct wth_object *)wthp_compositor)->id, ((struct wth_object *)id)->id);
  START_TIMING();
  ((struct wthp_compositor_interface *)(((struct wth_object *)wthp_compositor)->vfunc))->create_region
    (wthp_compositor, id);
  END_TIMING("wthp_compositor_create_region");
}

/* Function: wthp_blob_factory_create_buffer
 * Opcode 14 (hexadecimal: 0xe)
 */
static void function_14(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_blob_factory *wthp_blob_factory = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_buffer *buffer = (struct wthp_buffer *) wth_object_new_with_id (((struct wth_object *)wthp_blob_factory)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));
  unsigned int data_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  void *data = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int));
  int32_t *width = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int) + PADDED (data_sz) );
  int32_t *height = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int) + PADDED (data_sz)  + PADDED (sizeof (int32_t)));
  int32_t *stride = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int) + PADDED (data_sz)  + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));
  uint32_t *format = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + sizeof (unsigned int) + PADDED (data_sz)  + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));

  wth_debug ("wthp_blob_factory_create_buffer(%u, %u, [variable type uint32_t], [variable type void *], %d, %d, %d, %u) (opcode 14) called.", ((struct wth_object *)wthp_blob_factory)->id, ((struct wth_object *)buffer)->id, *width, *height, *stride, *format);
  START_TIMING();
  ((struct wthp_blob_factory_interface *)(((struct wth_object *)wthp_blob_factory)->vfunc))->create_buffer
    (wthp_blob_factory, buffer, data_sz, data, *width, *height, *stride, *format);
  END_TIMING("wthp_blob_factory_create_buffer");
}

/* Function: wthp_buffer_destroy
 * Opcode 16 (hexadecimal: 0x10)
 */
static void function_16(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_buffer *wthp_buffer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_buffer_destroy(%u) (opcode 16) called.", ((struct wth_object *)wthp_buffer)->id);
  START_TIMING();
  ((struct wthp_buffer_interface *)(((struct wth_object *)wthp_buffer)->vfunc))->destroy
    (wthp_buffer);
  END_TIMING("wthp_buffer_destroy");
}

/* Function: wthp_surface_destroy
 * Opcode 18 (hexadecimal: 0x12)
 */
static void function_18(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_surface_destroy(%u) (opcode 18) called.", ((struct wth_object *)wthp_surface)->id);
  START_TIMING();
  ((struct wthp_surface_interface *)(((struct wth_object *)wthp_surface)->vfunc))->destroy
    (wthp_surface);
  END_TIMING("wthp_surface_destroy");
}

/* Function: wthp_surface_attach
 * Opcode 19 (hexadecimal: 0x13)
 */
static void function_19(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_buffer *buffer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));
  int32_t *x = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  int32_t *y = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)));

  wth_debug ("wthp_surface_attach(%u, %u, %d, %d) (opcode 19) called.", ((struct wth_object *)wthp_surface)->id, ((struct wth_object *)buffer)->id, *x, *y);
  START_TIMING();
  ((struct wthp_surface_interface *)(((struct wth_object *)wthp_surface)->vfunc))->attach
    (wthp_surface, buffer, *x, *y);
  END_TIMING("wthp_surface_attach");
}

/* Function: wthp_surface_damage
 * Opcode 20 (hexadecimal: 0x14)
 */
static void function_20(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  int32_t *x = (void*)(body + PADDED (sizeof (uint32_t)));
  int32_t *y = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)));
  int32_t *width = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));
  int32_t *height = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));

  wth_debug ("wthp_surface_damage(%u, %d, %d, %d, %d) (opcode 20) called.", ((struct wth_object *)wthp_surface)->id, *x, *y, *width, *height);
  START_TIMING();
  ((struct wthp_surface_interface *)(((struct wth_object *)wthp_surface)->vfunc))->damage
    (wthp_surface, *x, *y, *width, *height);
  END_TIMING("wthp_surface_damage");
}

/* Function: wthp_surface_frame
 * Opcode 21 (hexadecimal: 0x15)
 */
static void function_21(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_callback *callback = (struct wthp_callback *) wth_object_new_with_id (((struct wth_object *)wthp_surface)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_surface_frame(%u, %u) (opcode 21) called.", ((struct wth_object *)wthp_surface)->id, ((struct wth_object *)callback)->id);
  START_TIMING();
  ((struct wthp_surface_interface *)(((struct wth_object *)wthp_surface)->vfunc))->frame
    (wthp_surface, callback);
  END_TIMING("wthp_surface_frame");
}

/* Function: wthp_surface_set_opaque_region
 * Opcode 22 (hexadecimal: 0x16)
 */
static void function_22(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_region *region = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_surface_set_opaque_region(%u, %u) (opcode 22) called.", ((struct wth_object *)wthp_surface)->id, ((struct wth_object *)region)->id);
  START_TIMING();
  ((struct wthp_surface_interface *)(((struct wth_object *)wthp_surface)->vfunc))->set_opaque_region
    (wthp_surface, region);
  END_TIMING("wthp_surface_set_opaque_region");
}

/* Function: wthp_surface_set_input_region
 * Opcode 23 (hexadecimal: 0x17)
 */
static void function_23(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_region *region = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_surface_set_input_region(%u, %u) (opcode 23) called.", ((struct wth_object *)wthp_surface)->id, ((struct wth_object *)region)->id);
  START_TIMING();
  ((struct wthp_surface_interface *)(((struct wth_object *)wthp_surface)->vfunc))->set_input_region
    (wthp_surface, region);
  END_TIMING("wthp_surface_set_input_region");
}

/* Function: wthp_surface_commit
 * Opcode 24 (hexadecimal: 0x18)
 */
static void function_24(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_surface_commit(%u) (opcode 24) called.", ((struct wth_object *)wthp_surface)->id);
  START_TIMING();
  ((struct wthp_surface_interface *)(((struct wth_object *)wthp_surface)->vfunc))->commit
    (wthp_surface);
  END_TIMING("wthp_surface_commit");
}

/* Function: wthp_surface_set_buffer_transform
 * Opcode 27 (hexadecimal: 0x1b)
 */
static void function_27(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  int32_t *transform = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_surface_set_buffer_transform(%u, %d) (opcode 27) called.", ((struct wth_object *)wthp_surface)->id, *transform);
  START_TIMING();
  ((struct wthp_surface_interface *)(((struct wth_object *)wthp_surface)->vfunc))->set_buffer_transform
    (wthp_surface, *transform);
  END_TIMING("wthp_surface_set_buffer_transform");
}

/* Function: wthp_surface_set_buffer_scale
 * Opcode 28 (hexadecimal: 0x1c)
 */
static void function_28(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  int32_t *scale = (void*)(body + PADDED (sizeof (uint32_t)));

  wth_debug ("wthp_surface_set_buffer_scale(%u, %d) (opcode 28) called.", ((struct wth_object *)wthp_surface)->id, *scale);
  START_TIMING();
  ((struct wthp_surface_interface *)(((struct wth_object *)wthp_surface)->vfunc))->set_buffer_scale
    (wthp_surface, *scale);
  END_TIMING("wthp_surface_set_buffer_scale");
}

/* Function: wthp_surface_damage_buffer
 * Opcode 29 (hexadecimal: 0x1d)
 */
static void function_29(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_surface *wthp_surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  int32_t *x = (void*)(body + PADDED (sizeof (uint32_t)));
  int32_t *y = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)));
  int32_t *width = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));
  int32_t *height = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));

  wth_debug ("wthp_surface_damage_buffer(%u, %d, %d, %d, %d) (opcode 29) called.", ((struct wth_object *)wthp_surface)->id, *x, *y, *width, *height);
  START_TIMING();
  ((struct wthp_surface_interface *)(((struct wth_object *)wthp_surface)->vfunc))->damage_buffer
    (wthp_surface, *x, *y, *width, *height);
  END_TIMING("wthp_surface_damage_buffer");
}

/* Function: wthp_seat_get_pointer
 * Opcode 31 (hexadecimal: 0x1f)
 */
static void function_31(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_seat *wthp_seat = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_pointer *id = (struct wthp_pointer *) wth_object_new_with_id (((struct wth_object *)wthp_seat)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_seat_get_pointer(%u, %u) (opcode 31) called.", ((struct wth_object *)wthp_seat)->id, ((struct wth_object *)id)->id);
  START_TIMING();
  ((struct wthp_seat_interface *)(((struct wth_object *)wthp_seat)->vfunc))->get_pointer
    (wthp_seat, id);
  END_TIMING("wthp_seat_get_pointer");
}

/* Function: wthp_seat_get_keyboard
 * Opcode 32 (hexadecimal: 0x20)
 */
static void function_32(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_seat *wthp_seat = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_keyboard *id = (struct wthp_keyboard *) wth_object_new_with_id (((struct wth_object *)wthp_seat)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_seat_get_keyboard(%u, %u) (opcode 32) called.", ((struct wth_object *)wthp_seat)->id, ((struct wth_object *)id)->id);
  START_TIMING();
  ((struct wthp_seat_interface *)(((struct wth_object *)wthp_seat)->vfunc))->get_keyboard
    (wthp_seat, id);
  END_TIMING("wthp_seat_get_keyboard");
}

/* Function: wthp_seat_get_touch
 * Opcode 33 (hexadecimal: 0x21)
 */
static void function_33(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_seat *wthp_seat = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  struct wthp_touch *id = (struct wthp_touch *) wth_object_new_with_id (((struct wth_object *)wthp_seat)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_seat_get_touch(%u, %u) (opcode 33) called.", ((struct wth_object *)wthp_seat)->id, ((struct wth_object *)id)->id);
  START_TIMING();
  ((struct wthp_seat_interface *)(((struct wth_object *)wthp_seat)->vfunc))->get_touch
    (wthp_seat, id);
  END_TIMING("wthp_seat_get_touch");
}

/* Function: wthp_seat_release
 * Opcode 35 (hexadecimal: 0x23)
 */
static void function_35(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_seat *wthp_seat = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_seat_release(%u) (opcode 35) called.", ((struct wth_object *)wthp_seat)->id);
  START_TIMING();
  ((struct wthp_seat_interface *)(((struct wth_object *)wthp_seat)->vfunc))->release
    (wthp_seat);
  END_TIMING("wthp_seat_release");
}

/* Function: wthp_pointer_set_cursor
 * Opcode 36 (hexadecimal: 0x24)
 */
static void function_36(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_pointer *wthp_pointer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *serial = (void*)(body + PADDED (sizeof (uint32_t)));
  struct wthp_surface *surface = (void *) wth_connection_get_object (conn, *(uint32_t *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t))));
  int32_t *hotspot_x = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)));
  int32_t *hotspot_y = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)));

  wth_debug ("wthp_pointer_set_cursor(%u, %u, %u, %d, %d) (opcode 36) called.", ((struct wth_object *)wthp_pointer)->id, *serial, ((struct wth_object *)surface)->id, *hotspot_x, *hotspot_y);
  START_TIMING();
  ((struct wthp_pointer_interface *)(((struct wth_object *)wthp_pointer)->vfunc))->set_cursor
    (wthp_pointer, *serial, surface, *hotspot_x, *hotspot_y);
  END_TIMING("wthp_pointer_set_cursor");
}

/* Function: wthp_pointer_release
 * Opcode 42 (hexadecimal: 0x2a)
 */
static void function_42(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_pointer *wthp_pointer = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_pointer_release(%u) (opcode 42) called.", ((struct wth_object *)wthp_pointer)->id);
  START_TIMING();
  ((struct wthp_pointer_interface *)(((struct wth_object *)wthp_pointer)->vfunc))->release
    (wthp_pointer);
  END_TIMING("wthp_pointer_release");
}

/* Function: wthp_keyboard_release
 * Opcode 52 (hexadecimal: 0x34)
 */
static void function_52(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_keyboard *wthp_keyboard = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_keyboard_release(%u) (opcode 52) called.", ((struct wth_object *)wthp_keyboard)->id);
  START_TIMING();
  ((struct wthp_keyboard_interface *)(((struct wth_object *)wthp_keyboard)->vfunc))->release
    (wthp_keyboard);
  END_TIMING("wthp_keyboard_release");
}

/* Function: wthp_touch_release
 * Opcode 59 (hexadecimal: 0x3b)
 */
static void function_59(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_touch *wthp_touch = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_touch_release(%u) (opcode 59) called.", ((struct wth_object *)wthp_touch)->id);
  START_TIMING();
  ((struct wthp_touch_interface *)(((struct wth_object *)wthp_touch)->vfunc))->release
    (wthp_touch);
  END_TIMING("wthp_touch_release");
}

/* Function: wthp_region_destroy
 * Opcode 64 (hexadecimal: 0x40)
 */
static void function_64(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_region *wthp_region = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_region_destroy(%u) (opcode 64) called.", ((struct wth_object *)wthp_region)->id);
  START_TIMING();
  ((struct wthp_region_interface *)(((struct wth_object *)wthp_region)->vfunc))->destroy
    (wthp_region);
  END_TIMING("wthp_region_destroy");
}

/* Function: wthp_region_add
 * Opcode 65 (hexadecimal: 0x41)
 */
static void function_65(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_region *wthp_region = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  int32_t *x = (void*)(body + PADDED (sizeof (uint32_t)));
  int32_t *y = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)));
  int32_t *width = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));
  int32_t *height = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));

  wth_debug ("wthp_region_add(%u, %d, %d, %d, %d) (opcode 65) called.", ((struct wth_object *)wthp_region)->id, *x, *y, *width, *height);
  START_TIMING();
  ((struct wthp_region_interface *)(((struct wth_object *)wthp_region)->vfunc))->add
    (wthp_region, *x, *y, *width, *height);
  END_TIMING("wthp_region_add");
}

/* Function: wthp_region_subtract
 * Opcode 66 (hexadecimal: 0x42)
 */
static void function_66(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_region *wthp_region = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  int32_t *x = (void*)(body + PADDED (sizeof (uint32_t)));
  int32_t *y = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)));
  int32_t *width = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));
  int32_t *height = (void*)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)) + PADDED (sizeof (int32_t)));

  wth_debug ("wthp_region_subtract(%u, %d, %d, %d, %d) (opcode 66) called.", ((struct wth_object *)wthp_region)->id, *x, *y, *width, *height);
  START_TIMING();
  ((struct wthp_region_interface *)(((struct wth_object *)wthp_region)->vfunc))->subtract
    (wthp_region, *x, *y, *width, *height);
  END_TIMING("wthp_region_subtract");
}

/* Function: wthp_farstream_destroy
 * Opcode 67 (hexadecimal: 0x43)
 */
static void function_67(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_farstream *wthp_farstream = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_farstream_destroy(%u) (opcode 67) called.", ((struct wth_object *)wthp_farstream)->id);
  START_TIMING();
  ((struct wthp_farstream_interface *)(((struct wth_object *)wthp_farstream)->vfunc))->destroy
    (wthp_farstream);
  END_TIMING("wthp_farstream_destroy");
}

/* Function: wthp_farstream_connect
 * Opcode 68 (hexadecimal: 0x44)
 */
static void function_68(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_farstream *wthp_farstream = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *port = (void*)(body + PADDED (sizeof (uint32_t)));
  struct wthp_farstream_remote *remote = (struct wthp_farstream_remote *) wth_object_new_with_id (((struct wth_object *)wthp_farstream)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_farstream_connect(%u, %u, %u) (opcode 68) called.", ((struct wth_object *)wthp_farstream)->id, *port, ((struct wth_object *)remote)->id);
  START_TIMING();
  ((struct wthp_farstream_interface *)(((struct wth_object *)wthp_farstream)->vfunc))->connect
    (wthp_farstream, *port, remote);
  END_TIMING("wthp_farstream_connect");
}

/* Function: wthp_farstream_remote_destroy
 * Opcode 69 (hexadecimal: 0x45)
 */
static void function_69(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_farstream_remote *wthp_farstream_remote = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));

  wth_debug ("wthp_farstream_remote_destroy(%u) (opcode 69) called.", ((struct wth_object *)wthp_farstream_remote)->id);
  START_TIMING();
  ((struct wthp_farstream_remote_interface *)(((struct wth_object *)wthp_farstream_remote)->vfunc))->destroy
    (wthp_farstream_remote);
  END_TIMING("wthp_farstream_remote_destroy");
}

/* Function: wthp_farstream_remote_codec_offer
 * Opcode 70 (hexadecimal: 0x46)
 */
static void function_70(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_farstream_remote *wthp_farstream_remote = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  unsigned int list_sz __attribute__((unused)) = *(unsigned int *)(body + PADDED (sizeof (uint32_t)));
  const char *list = (void*)(body + PADDED (sizeof (uint32_t)) + sizeof (unsigned int));

  wth_debug ("wthp_farstream_remote_codec_offer(%u, [variable type const char *]) (opcode 70) called.", ((struct wth_object *)wthp_farstream_remote)->id);
  START_TIMING();
  ((struct wthp_farstream_remote_interface *)(((struct wth_object *)wthp_farstream_remote)->vfunc))->codec_offer
    (wthp_farstream_remote, list);
  END_TIMING("wthp_farstream_remote_codec_offer");
}

/* Function: wthp_farstream_remote_create_buffer
 * Opcode 71 (hexadecimal: 0x47)
 */
static void function_71(struct wth_connection *conn,
    const hdr_t *header, const char *body)
{
  struct wthp_farstream_remote *wthp_farstream_remote = (void *) wth_connection_get_object (conn, *(uint32_t *)(body));
  uint32_t *frame_id = (void*)(body + PADDED (sizeof (uint32_t)));
  struct wthp_buffer *id = (struct wthp_buffer *) wth_object_new_with_id (((struct wth_object *)wthp_farstream_remote)->connection, *(uint32_t *)(body + PADDED (sizeof (uint32_t)) + PADDED (sizeof (uint32_t))));

  wth_debug ("wthp_farstream_remote_create_buffer(%u, %u, %u) (opcode 71) called.", ((struct wth_object *)wthp_farstream_remote)->id, *frame_id, ((struct wth_object *)id)->id);
  START_TIMING();
  ((struct wthp_farstream_remote_interface *)(((struct wth_object *)wthp_farstream_remote)->vfunc))->create_buffer
    (wthp_farstream_remote, *frame_id, id);
  END_TIMING("wthp_farstream_remote_create_buffer");
}

const int demarshaller__max_opcode = 74;
const demarshaller_helper_function_t request_demarshaller_functions[] = {
  NULL,
  function_1,
  function_2,
  function_3,
  NULL,
  NULL,
  NULL,
  function_7,
  function_8,
  NULL,
  NULL,
  NULL,
  function_12,
  function_13,
  function_14,
  NULL,
  function_16,
  NULL,
  function_18,
  function_19,
  function_20,
  function_21,
  function_22,
  function_23,
  function_24,
  NULL,
  NULL,
  function_27,
  function_28,
  function_29,
  NULL,
  function_31,
  function_32,
  function_33,
  NULL,
  function_35,
  function_36,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  function_42,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  function_52,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  function_59,
  NULL,
  NULL,
  NULL,
  NULL,
  function_64,
  function_65,
  function_66,
  function_67,
  function_68,
  function_69,
  function_70,
  function_71,
  NULL,
  NULL,
  NULL,
};
