/*
 * Copyright © 2013-2014 Collabora, Ltd.
 * Copyright © 2016 DENSO CORPORATION
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "waltham-object.h"
#include "waltham-connection.h"
#include "waltham-private.h"

#include "marshaller_log.h"
#include "marshaller.h"

#include "waltham-server.h"

WTH_EXPORT void
wth_display_send_error (struct wth_display * wth_display, struct wth_object * object_id, uint32_t code, const char * message)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   int message_sz = strlen (message) + 1;
   sz += sizeof (unsigned int) + PADDED(message_sz);

   START_TIMING("wth_display_send_error", "[return/variable/unprintable type struct wth_display *], [return/variable/unprintable type struct wth_object *], %u, %s", code, message);
   START_MESSAGE("wth_display_send_error", sz, 4);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wth_display)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)object_id)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&code, sizeof(code) );
   SERIALIZE_DATA( (void *)message, message_sz);
   END_MESSAGE(((struct wth_object *)wth_display)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wth_display_send_delete_id (struct wth_display * wth_display, uint32_t id)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wth_display_send_delete_id", "[return/variable/unprintable type struct wth_display *], %u", id);
   START_MESSAGE("wth_display_send_delete_id", sz, 5);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wth_display)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&id, sizeof(id) );
   END_MESSAGE(((struct wth_object *)wth_display)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wth_display_send_server_version (struct wth_display * wth_display, uint32_t server_version)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wth_display_send_server_version", "[return/variable/unprintable type struct wth_display *], %u", server_version);
   START_MESSAGE("wth_display_send_server_version", sz, 6);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wth_display)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&server_version, sizeof(server_version) );
   END_MESSAGE(((struct wth_object *)wth_display)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_registry_send_global (struct wthp_registry * wthp_registry, uint32_t name, const char * interface, uint32_t version)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   int interface_sz = strlen (interface) + 1;
   sz += sizeof (unsigned int) + PADDED(interface_sz);

   START_TIMING("wthp_registry_send_global", "[return/variable/unprintable type struct wthp_registry *], %u, %s, %u", name, interface, version);
   START_MESSAGE("wthp_registry_send_global", sz, 9);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_registry)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&name, sizeof(name) );
   SERIALIZE_DATA( (void *)interface, interface_sz);
   SERIALIZE_PARAM( (void *)&version, sizeof(version) );
   END_MESSAGE(((struct wth_object *)wthp_registry)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_registry_send_global_remove (struct wthp_registry * wthp_registry, uint32_t name)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_registry_send_global_remove", "[return/variable/unprintable type struct wthp_registry *], %u", name);
   START_MESSAGE("wthp_registry_send_global_remove", sz, 10);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_registry)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&name, sizeof(name) );
   END_MESSAGE(((struct wth_object *)wthp_registry)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_callback_send_done (struct wthp_callback * wthp_callback, uint32_t callback_data)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_callback_send_done", "[return/variable/unprintable type struct wthp_callback *], %u", callback_data);
   START_MESSAGE("wthp_callback_send_done", sz, 11);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_callback)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&callback_data, sizeof(callback_data) );
   END_MESSAGE(((struct wth_object *)wthp_callback)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_blob_factory_send_format (struct wthp_blob_factory * wthp_blob_factory, uint32_t format)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_blob_factory_send_format", "[return/variable/unprintable type struct wthp_blob_factory *], %u", format);
   START_MESSAGE("wthp_blob_factory_send_format", sz, 15);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_blob_factory)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&format, sizeof(format) );
   END_MESSAGE(((struct wth_object *)wthp_blob_factory)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_buffer_send_complete (struct wthp_buffer * wthp_buffer, uint32_t serial)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_buffer_send_complete", "[return/variable/unprintable type struct wthp_buffer *], %u", serial);
   START_MESSAGE("wthp_buffer_send_complete", sz, 17);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_buffer)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&serial, sizeof(serial) );
   END_MESSAGE(((struct wth_object *)wthp_buffer)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_surface_send_enter (struct wthp_surface * wthp_surface, struct wthp_output * output)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_surface_send_enter", "[return/variable/unprintable type struct wthp_surface *], [return/variable/unprintable type struct wthp_output *]");
   START_MESSAGE("wthp_surface_send_enter", sz, 25);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)output)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_surface_send_leave (struct wthp_surface * wthp_surface, struct wthp_output * output)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_surface_send_leave", "[return/variable/unprintable type struct wthp_surface *], [return/variable/unprintable type struct wthp_output *]");
   START_MESSAGE("wthp_surface_send_leave", sz, 26);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)output)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_surface)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_seat_send_capabilities (struct wthp_seat * wthp_seat, uint32_t capabilities)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_seat_send_capabilities", "[return/variable/unprintable type struct wthp_seat *], %u", capabilities);
   START_MESSAGE("wthp_seat_send_capabilities", sz, 30);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_seat)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&capabilities, sizeof(capabilities) );
   END_MESSAGE(((struct wth_object *)wthp_seat)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_seat_send_name (struct wthp_seat * wthp_seat, const char * name)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   int name_sz = strlen (name) + 1;
   sz += sizeof (unsigned int) + PADDED(name_sz);

   START_TIMING("wthp_seat_send_name", "[return/variable/unprintable type struct wthp_seat *], %s", name);
   START_MESSAGE("wthp_seat_send_name", sz, 34);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_seat)->id, sizeof(uint32_t) );
   SERIALIZE_DATA( (void *)name, name_sz);
   END_MESSAGE(((struct wth_object *)wthp_seat)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_pointer_send_enter (struct wthp_pointer * wthp_pointer, uint32_t serial, struct wthp_surface * surface, wth_fixed_t surface_x, wth_fixed_t surface_y)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(wth_fixed_t)) + PADDED(sizeof(wth_fixed_t));

   START_TIMING("wthp_pointer_send_enter", "[return/variable/unprintable type struct wthp_pointer *], %u, [return/variable/unprintable type struct wthp_surface *], [return/variable/unprintable type wth_fixed_t], [return/variable/unprintable type wth_fixed_t]", serial);
   START_MESSAGE("wthp_pointer_send_enter", sz, 37);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_pointer)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&serial, sizeof(serial) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&surface_x, sizeof(surface_x) );
   SERIALIZE_PARAM( (void *)&surface_y, sizeof(surface_y) );
   END_MESSAGE(((struct wth_object *)wthp_pointer)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_pointer_send_leave (struct wthp_pointer * wthp_pointer, uint32_t serial, struct wthp_surface * surface)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_pointer_send_leave", "[return/variable/unprintable type struct wthp_pointer *], %u, [return/variable/unprintable type struct wthp_surface *]", serial);
   START_MESSAGE("wthp_pointer_send_leave", sz, 38);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_pointer)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&serial, sizeof(serial) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)surface)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_pointer)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_pointer_send_motion (struct wthp_pointer * wthp_pointer, uint32_t time, wth_fixed_t surface_x, wth_fixed_t surface_y)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(wth_fixed_t)) + PADDED(sizeof(wth_fixed_t));

   START_TIMING("wthp_pointer_send_motion", "[return/variable/unprintable type struct wthp_pointer *], %u, [return/variable/unprintable type wth_fixed_t], [return/variable/unprintable type wth_fixed_t]", time);
   START_MESSAGE("wthp_pointer_send_motion", sz, 39);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_pointer)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&time, sizeof(time) );
   SERIALIZE_PARAM( (void *)&surface_x, sizeof(surface_x) );
   SERIALIZE_PARAM( (void *)&surface_y, sizeof(surface_y) );
   END_MESSAGE(((struct wth_object *)wthp_pointer)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_pointer_send_button (struct wthp_pointer * wthp_pointer, uint32_t serial, uint32_t time, uint32_t button, uint32_t state)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_pointer_send_button", "[return/variable/unprintable type struct wthp_pointer *], %u, %u, %u, %u", serial, time, button, state);
   START_MESSAGE("wthp_pointer_send_button", sz, 40);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_pointer)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&serial, sizeof(serial) );
   SERIALIZE_PARAM( (void *)&time, sizeof(time) );
   SERIALIZE_PARAM( (void *)&button, sizeof(button) );
   SERIALIZE_PARAM( (void *)&state, sizeof(state) );
   END_MESSAGE(((struct wth_object *)wthp_pointer)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_pointer_send_axis (struct wthp_pointer * wthp_pointer, uint32_t time, uint32_t axis, wth_fixed_t value)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(wth_fixed_t));

   START_TIMING("wthp_pointer_send_axis", "[return/variable/unprintable type struct wthp_pointer *], %u, %u, [return/variable/unprintable type wth_fixed_t]", time, axis);
   START_MESSAGE("wthp_pointer_send_axis", sz, 41);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_pointer)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&time, sizeof(time) );
   SERIALIZE_PARAM( (void *)&axis, sizeof(axis) );
   SERIALIZE_PARAM( (void *)&value, sizeof(value) );
   END_MESSAGE(((struct wth_object *)wthp_pointer)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_pointer_send_frame (struct wthp_pointer * wthp_pointer)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_pointer_send_frame", "[return/variable/unprintable type struct wthp_pointer *]");
   START_MESSAGE("wthp_pointer_send_frame", sz, 43);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_pointer)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_pointer)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_pointer_send_axis_source (struct wthp_pointer * wthp_pointer, uint32_t axis_source)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_pointer_send_axis_source", "[return/variable/unprintable type struct wthp_pointer *], %u", axis_source);
   START_MESSAGE("wthp_pointer_send_axis_source", sz, 44);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_pointer)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&axis_source, sizeof(axis_source) );
   END_MESSAGE(((struct wth_object *)wthp_pointer)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_pointer_send_axis_stop (struct wthp_pointer * wthp_pointer, uint32_t time, uint32_t axis)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_pointer_send_axis_stop", "[return/variable/unprintable type struct wthp_pointer *], %u, %u", time, axis);
   START_MESSAGE("wthp_pointer_send_axis_stop", sz, 45);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_pointer)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&time, sizeof(time) );
   SERIALIZE_PARAM( (void *)&axis, sizeof(axis) );
   END_MESSAGE(((struct wth_object *)wthp_pointer)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_pointer_send_axis_discrete (struct wthp_pointer * wthp_pointer, uint32_t axis, int32_t discrete)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_pointer_send_axis_discrete", "[return/variable/unprintable type struct wthp_pointer *], %u, %d", axis, discrete);
   START_MESSAGE("wthp_pointer_send_axis_discrete", sz, 46);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_pointer)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&axis, sizeof(axis) );
   SERIALIZE_PARAM( (void *)&discrete, sizeof(discrete) );
   END_MESSAGE(((struct wth_object *)wthp_pointer)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_keyboard_send_keymap (struct wthp_keyboard * wthp_keyboard, uint32_t format, uint32_t keymap_sz, void * keymap)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(keymap_sz);

   START_TIMING("wthp_keyboard_send_keymap", "[return/variable/unprintable type struct wthp_keyboard *], %u, %u, [return/variable/unprintable type void *]", format, keymap_sz);
   START_MESSAGE("wthp_keyboard_send_keymap", sz, 47);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_keyboard)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&format, sizeof(format) );
   SERIALIZE_DATA( (void *)keymap, keymap_sz);
   END_MESSAGE(((struct wth_object *)wthp_keyboard)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_keyboard_send_enter (struct wthp_keyboard * wthp_keyboard, uint32_t serial, struct wthp_surface * surface, struct wth_array * keys)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(struct wth_array *));

   START_TIMING("wthp_keyboard_send_enter", "[return/variable/unprintable type struct wthp_keyboard *], %u, [return/variable/unprintable type struct wthp_surface *], [return/variable/unprintable type struct wth_array *]", serial);
   START_MESSAGE("wthp_keyboard_send_enter", sz, 48);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_keyboard)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&serial, sizeof(serial) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)surface)->id, sizeof(uint32_t) );
   SERIALIZE_ARRAY( keys );
   END_MESSAGE(((struct wth_object *)wthp_keyboard)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_keyboard_send_leave (struct wthp_keyboard * wthp_keyboard, uint32_t serial, struct wthp_surface * surface)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_keyboard_send_leave", "[return/variable/unprintable type struct wthp_keyboard *], %u, [return/variable/unprintable type struct wthp_surface *]", serial);
   START_MESSAGE("wthp_keyboard_send_leave", sz, 49);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_keyboard)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&serial, sizeof(serial) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)surface)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_keyboard)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_keyboard_send_key (struct wthp_keyboard * wthp_keyboard, uint32_t serial, uint32_t time, uint32_t key, uint32_t state)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_keyboard_send_key", "[return/variable/unprintable type struct wthp_keyboard *], %u, %u, %u, %u", serial, time, key, state);
   START_MESSAGE("wthp_keyboard_send_key", sz, 50);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_keyboard)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&serial, sizeof(serial) );
   SERIALIZE_PARAM( (void *)&time, sizeof(time) );
   SERIALIZE_PARAM( (void *)&key, sizeof(key) );
   SERIALIZE_PARAM( (void *)&state, sizeof(state) );
   END_MESSAGE(((struct wth_object *)wthp_keyboard)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_keyboard_send_modifiers (struct wthp_keyboard * wthp_keyboard, uint32_t serial, uint32_t mods_depressed, uint32_t mods_latched, uint32_t mods_locked, uint32_t group)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_keyboard_send_modifiers", "[return/variable/unprintable type struct wthp_keyboard *], %u, %u, %u, %u, %u", serial, mods_depressed, mods_latched, mods_locked, group);
   START_MESSAGE("wthp_keyboard_send_modifiers", sz, 51);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_keyboard)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&serial, sizeof(serial) );
   SERIALIZE_PARAM( (void *)&mods_depressed, sizeof(mods_depressed) );
   SERIALIZE_PARAM( (void *)&mods_latched, sizeof(mods_latched) );
   SERIALIZE_PARAM( (void *)&mods_locked, sizeof(mods_locked) );
   SERIALIZE_PARAM( (void *)&group, sizeof(group) );
   END_MESSAGE(((struct wth_object *)wthp_keyboard)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_keyboard_send_repeat_info (struct wthp_keyboard * wthp_keyboard, int32_t rate, int32_t delay)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_keyboard_send_repeat_info", "[return/variable/unprintable type struct wthp_keyboard *], %d, %d", rate, delay);
   START_MESSAGE("wthp_keyboard_send_repeat_info", sz, 53);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_keyboard)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&rate, sizeof(rate) );
   SERIALIZE_PARAM( (void *)&delay, sizeof(delay) );
   END_MESSAGE(((struct wth_object *)wthp_keyboard)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_touch_send_down (struct wthp_touch * wthp_touch, uint32_t serial, uint32_t time, struct wthp_surface * surface, int32_t id, wth_fixed_t x, wth_fixed_t y)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(wth_fixed_t)) + PADDED(sizeof(wth_fixed_t));

   START_TIMING("wthp_touch_send_down", "[return/variable/unprintable type struct wthp_touch *], %u, %u, [return/variable/unprintable type struct wthp_surface *], %d, [return/variable/unprintable type wth_fixed_t], [return/variable/unprintable type wth_fixed_t]", serial, time, id);
   START_MESSAGE("wthp_touch_send_down", sz, 54);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_touch)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&serial, sizeof(serial) );
   SERIALIZE_PARAM( (void *)&time, sizeof(time) );
   SERIALIZE_PARAM( (void *)&((struct wth_object *)surface)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&id, sizeof(id) );
   SERIALIZE_PARAM( (void *)&x, sizeof(x) );
   SERIALIZE_PARAM( (void *)&y, sizeof(y) );
   END_MESSAGE(((struct wth_object *)wthp_touch)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_touch_send_up (struct wthp_touch * wthp_touch, uint32_t serial, uint32_t time, int32_t id)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_touch_send_up", "[return/variable/unprintable type struct wthp_touch *], %u, %u, %d", serial, time, id);
   START_MESSAGE("wthp_touch_send_up", sz, 55);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_touch)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&serial, sizeof(serial) );
   SERIALIZE_PARAM( (void *)&time, sizeof(time) );
   SERIALIZE_PARAM( (void *)&id, sizeof(id) );
   END_MESSAGE(((struct wth_object *)wthp_touch)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_touch_send_motion (struct wthp_touch * wthp_touch, uint32_t time, int32_t id, wth_fixed_t x, wth_fixed_t y)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(wth_fixed_t)) + PADDED(sizeof(wth_fixed_t));

   START_TIMING("wthp_touch_send_motion", "[return/variable/unprintable type struct wthp_touch *], %u, %d, [return/variable/unprintable type wth_fixed_t], [return/variable/unprintable type wth_fixed_t]", time, id);
   START_MESSAGE("wthp_touch_send_motion", sz, 56);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_touch)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&time, sizeof(time) );
   SERIALIZE_PARAM( (void *)&id, sizeof(id) );
   SERIALIZE_PARAM( (void *)&x, sizeof(x) );
   SERIALIZE_PARAM( (void *)&y, sizeof(y) );
   END_MESSAGE(((struct wth_object *)wthp_touch)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_touch_send_frame (struct wthp_touch * wthp_touch)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_touch_send_frame", "[return/variable/unprintable type struct wthp_touch *]");
   START_MESSAGE("wthp_touch_send_frame", sz, 57);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_touch)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_touch)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_touch_send_cancel (struct wthp_touch * wthp_touch)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_touch_send_cancel", "[return/variable/unprintable type struct wthp_touch *]");
   START_MESSAGE("wthp_touch_send_cancel", sz, 58);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_touch)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_touch)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_output_send_geometry (struct wthp_output * wthp_output, int32_t x, int32_t y, int32_t physical_width, int32_t physical_height, int32_t subpixel, const char * make, const char * model, int32_t transform)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t));

   int make_sz = strlen (make) + 1;
   sz += sizeof (unsigned int) + PADDED(make_sz);
   int model_sz = strlen (model) + 1;
   sz += sizeof (unsigned int) + PADDED(model_sz);

   START_TIMING("wthp_output_send_geometry", "[return/variable/unprintable type struct wthp_output *], %d, %d, %d, %d, %d, %s, %s, %d", x, y, physical_width, physical_height, subpixel, make, model, transform);
   START_MESSAGE("wthp_output_send_geometry", sz, 60);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_output)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&x, sizeof(x) );
   SERIALIZE_PARAM( (void *)&y, sizeof(y) );
   SERIALIZE_PARAM( (void *)&physical_width, sizeof(physical_width) );
   SERIALIZE_PARAM( (void *)&physical_height, sizeof(physical_height) );
   SERIALIZE_PARAM( (void *)&subpixel, sizeof(subpixel) );
   SERIALIZE_DATA( (void *)make, make_sz);
   SERIALIZE_DATA( (void *)model, model_sz);
   SERIALIZE_PARAM( (void *)&transform, sizeof(transform) );
   END_MESSAGE(((struct wth_object *)wthp_output)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_output_send_mode (struct wthp_output * wthp_output, uint32_t flags, int32_t width, int32_t height, int32_t refresh)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_output_send_mode", "[return/variable/unprintable type struct wthp_output *], %u, %d, %d, %d", flags, width, height, refresh);
   START_MESSAGE("wthp_output_send_mode", sz, 61);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_output)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&flags, sizeof(flags) );
   SERIALIZE_PARAM( (void *)&width, sizeof(width) );
   SERIALIZE_PARAM( (void *)&height, sizeof(height) );
   SERIALIZE_PARAM( (void *)&refresh, sizeof(refresh) );
   END_MESSAGE(((struct wth_object *)wthp_output)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_output_send_done (struct wthp_output * wthp_output)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_output_send_done", "[return/variable/unprintable type struct wthp_output *]");
   START_MESSAGE("wthp_output_send_done", sz, 62);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_output)->id, sizeof(uint32_t) );
   END_MESSAGE(((struct wth_object *)wthp_output)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_output_send_scale (struct wthp_output * wthp_output, int32_t factor)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(int32_t));

   START_TIMING("wthp_output_send_scale", "[return/variable/unprintable type struct wthp_output *], %d", factor);
   START_MESSAGE("wthp_output_send_scale", sz, 63);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_output)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&factor, sizeof(factor) );
   END_MESSAGE(((struct wth_object *)wthp_output)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_farstream_remote_send_port (struct wthp_farstream_remote * wthp_farstream_remote, uint32_t port)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   START_TIMING("wthp_farstream_remote_send_port", "[return/variable/unprintable type struct wthp_farstream_remote *], %u", port);
   START_MESSAGE("wthp_farstream_remote_send_port", sz, 72);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_farstream_remote)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&port, sizeof(port) );
   END_MESSAGE(((struct wth_object *)wthp_farstream_remote)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_farstream_remote_send_codec_answer (struct wthp_farstream_remote * wthp_farstream_remote, const char * list)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t));

   int list_sz = strlen (list) + 1;
   sz += sizeof (unsigned int) + PADDED(list_sz);

   START_TIMING("wthp_farstream_remote_send_codec_answer", "[return/variable/unprintable type struct wthp_farstream_remote *], %s", list);
   START_MESSAGE("wthp_farstream_remote_send_codec_answer", sz, 73);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_farstream_remote)->id, sizeof(uint32_t) );
   SERIALIZE_DATA( (void *)list, list_sz);
   END_MESSAGE(((struct wth_object *)wthp_farstream_remote)->connection);
   END_TIMING("");
}

WTH_EXPORT void
wthp_farstream_remote_send_error (struct wthp_farstream_remote * wthp_farstream_remote, uint32_t code, const char * message)
{
   int sz = sizeof(hdr_t) + PADDED(sizeof(uint32_t)) + PADDED(sizeof(uint32_t));

   int message_sz = strlen (message) + 1;
   sz += sizeof (unsigned int) + PADDED(message_sz);

   START_TIMING("wthp_farstream_remote_send_error", "[return/variable/unprintable type struct wthp_farstream_remote *], %u, %s", code, message);
   START_MESSAGE("wthp_farstream_remote_send_error", sz, 74);
   SERIALIZE_PARAM( (void *)&((struct wth_object *)wthp_farstream_remote)->id, sizeof(uint32_t) );
   SERIALIZE_PARAM( (void *)&code, sizeof(code) );
   SERIALIZE_DATA( (void *)message, message_sz);
   END_MESSAGE(((struct wth_object *)wthp_farstream_remote)->connection);
   END_TIMING("");
}

