/*
 * Copyright © 2013-2014 Collabora, Ltd.
 * Copyright © 2016 DENSO CORPORATION
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <stdint.h>

#include <waltham-object.h>
#include <waltham-util.h>

#ifdef  __cplusplus
extern "C" {
#endif

struct wth_display;
struct wthp_blob_factory;
struct wthp_buffer;
struct wthp_callback;
struct wthp_compositor;
struct wthp_farstream;
struct wthp_farstream_remote;
struct wthp_keyboard;
struct wthp_output;
struct wthp_pointer;
struct wthp_region;
struct wthp_registry;
struct wthp_resource;
struct wthp_seat;
struct wthp_surface;
struct wthp_touch;

#ifndef WALTHAM_CLIENT_H
#define WALTHAM_CLIENT_H

struct wth_display_listener {
  void (*error) (struct wth_display * wth_display, struct wth_object * object_id, uint32_t code, const char * message);
  void (*delete_id) (struct wth_display * wth_display, uint32_t id);
  void (*server_version) (struct wth_display * wth_display, uint32_t server_version);
};

static inline void
wth_display_set_listener(struct wth_display *self, const struct wth_display_listener *funcs, void *user_data)
{
  wth_object_set_listener((struct wth_object *)self, (void (**)(void)) funcs, user_data);
}

struct wthp_registry_listener {
  void (*global) (struct wthp_registry * wthp_registry, uint32_t name, const char * interface, uint32_t version);
  void (*global_remove) (struct wthp_registry * wthp_registry, uint32_t name);
};

static inline void
wthp_registry_set_listener(struct wthp_registry *self, const struct wthp_registry_listener *funcs, void *user_data)
{
  wth_object_set_listener((struct wth_object *)self, (void (**)(void)) funcs, user_data);
}

struct wthp_callback_listener {
  void (*done) (struct wthp_callback * wthp_callback, uint32_t callback_data);
};

static inline void
wthp_callback_set_listener(struct wthp_callback *self, const struct wthp_callback_listener *funcs, void *user_data)
{
  wth_object_set_listener((struct wth_object *)self, (void (**)(void)) funcs, user_data);
}

struct wthp_blob_factory_listener {
  void (*format) (struct wthp_blob_factory * wthp_blob_factory, uint32_t format);
};

static inline void
wthp_blob_factory_set_listener(struct wthp_blob_factory *self, const struct wthp_blob_factory_listener *funcs, void *user_data)
{
  wth_object_set_listener((struct wth_object *)self, (void (**)(void)) funcs, user_data);
}

struct wthp_buffer_listener {
  void (*complete) (struct wthp_buffer * wthp_buffer, uint32_t serial);
};

static inline void
wthp_buffer_set_listener(struct wthp_buffer *self, const struct wthp_buffer_listener *funcs, void *user_data)
{
  wth_object_set_listener((struct wth_object *)self, (void (**)(void)) funcs, user_data);
}

struct wthp_surface_listener {
  void (*enter) (struct wthp_surface * wthp_surface, struct wthp_output * output);
  void (*leave) (struct wthp_surface * wthp_surface, struct wthp_output * output);
};

static inline void
wthp_surface_set_listener(struct wthp_surface *self, const struct wthp_surface_listener *funcs, void *user_data)
{
  wth_object_set_listener((struct wth_object *)self, (void (**)(void)) funcs, user_data);
}

struct wthp_seat_listener {
  void (*capabilities) (struct wthp_seat * wthp_seat, uint32_t capabilities);
  void (*name) (struct wthp_seat * wthp_seat, const char * name);
};

static inline void
wthp_seat_set_listener(struct wthp_seat *self, const struct wthp_seat_listener *funcs, void *user_data)
{
  wth_object_set_listener((struct wth_object *)self, (void (**)(void)) funcs, user_data);
}

struct wthp_pointer_listener {
  void (*enter) (struct wthp_pointer * wthp_pointer, uint32_t serial, struct wthp_surface * surface, wth_fixed_t surface_x, wth_fixed_t surface_y);
  void (*leave) (struct wthp_pointer * wthp_pointer, uint32_t serial, struct wthp_surface * surface);
  void (*motion) (struct wthp_pointer * wthp_pointer, uint32_t time, wth_fixed_t surface_x, wth_fixed_t surface_y);
  void (*button) (struct wthp_pointer * wthp_pointer, uint32_t serial, uint32_t time, uint32_t button, uint32_t state);
  void (*axis) (struct wthp_pointer * wthp_pointer, uint32_t time, uint32_t axis, wth_fixed_t value);
  void (*frame) (struct wthp_pointer * wthp_pointer);
  void (*axis_source) (struct wthp_pointer * wthp_pointer, uint32_t axis_source);
  void (*axis_stop) (struct wthp_pointer * wthp_pointer, uint32_t time, uint32_t axis);
  void (*axis_discrete) (struct wthp_pointer * wthp_pointer, uint32_t axis, int32_t discrete);
};

static inline void
wthp_pointer_set_listener(struct wthp_pointer *self, const struct wthp_pointer_listener *funcs, void *user_data)
{
  wth_object_set_listener((struct wth_object *)self, (void (**)(void)) funcs, user_data);
}

struct wthp_keyboard_listener {
  void (*keymap) (struct wthp_keyboard * wthp_keyboard, uint32_t format, uint32_t keymap_sz, void * keymap);
  void (*enter) (struct wthp_keyboard * wthp_keyboard, uint32_t serial, struct wthp_surface * surface, struct wth_array * keys);
  void (*leave) (struct wthp_keyboard * wthp_keyboard, uint32_t serial, struct wthp_surface * surface);
  void (*key) (struct wthp_keyboard * wthp_keyboard, uint32_t serial, uint32_t time, uint32_t key, uint32_t state);
  void (*modifiers) (struct wthp_keyboard * wthp_keyboard, uint32_t serial, uint32_t mods_depressed, uint32_t mods_latched, uint32_t mods_locked, uint32_t group);
  void (*repeat_info) (struct wthp_keyboard * wthp_keyboard, int32_t rate, int32_t delay);
};

static inline void
wthp_keyboard_set_listener(struct wthp_keyboard *self, const struct wthp_keyboard_listener *funcs, void *user_data)
{
  wth_object_set_listener((struct wth_object *)self, (void (**)(void)) funcs, user_data);
}

struct wthp_touch_listener {
  void (*down) (struct wthp_touch * wthp_touch, uint32_t serial, uint32_t time, struct wthp_surface * surface, int32_t id, wth_fixed_t x, wth_fixed_t y);
  void (*up) (struct wthp_touch * wthp_touch, uint32_t serial, uint32_t time, int32_t id);
  void (*motion) (struct wthp_touch * wthp_touch, uint32_t time, int32_t id, wth_fixed_t x, wth_fixed_t y);
  void (*frame) (struct wthp_touch * wthp_touch);
  void (*cancel) (struct wthp_touch * wthp_touch);
};

static inline void
wthp_touch_set_listener(struct wthp_touch *self, const struct wthp_touch_listener *funcs, void *user_data)
{
  wth_object_set_listener((struct wth_object *)self, (void (**)(void)) funcs, user_data);
}

struct wthp_output_listener {
  void (*geometry) (struct wthp_output * wthp_output, int32_t x, int32_t y, int32_t physical_width, int32_t physical_height, int32_t subpixel, const char * make, const char * model, int32_t transform);
  void (*mode) (struct wthp_output * wthp_output, uint32_t flags, int32_t width, int32_t height, int32_t refresh);
  void (*done) (struct wthp_output * wthp_output);
  void (*scale) (struct wthp_output * wthp_output, int32_t factor);
};

static inline void
wthp_output_set_listener(struct wthp_output *self, const struct wthp_output_listener *funcs, void *user_data)
{
  wth_object_set_listener((struct wth_object *)self, (void (**)(void)) funcs, user_data);
}

struct wthp_farstream_remote_listener {
  void (*port) (struct wthp_farstream_remote * wthp_farstream_remote, uint32_t port);
  void (*codec_answer) (struct wthp_farstream_remote * wthp_farstream_remote, const char * list);
  void (*error) (struct wthp_farstream_remote * wthp_farstream_remote, uint32_t code, const char * message);
};

static inline void
wth_display_free(struct wth_display *self)
{
  wth_object_delete((struct wth_object *)self);
}

void
wth_display_client_version (struct wth_display * wth_display, uint32_t client_version);

struct wthp_callback *
wth_display_sync (struct wth_display * wth_display);

struct wthp_registry *
wth_display_get_registry (struct wth_display * wth_display);

static inline void
wthp_registry_free(struct wthp_registry *self)
{
  wth_object_delete((struct wth_object *)self);
}

void
wthp_registry_destroy (struct wthp_registry * wthp_registry);

struct wth_object *
wthp_registry_bind (struct wthp_registry * wthp_registry, uint32_t name, const char * interface, uint32_t version);

static inline void
wthp_callback_free(struct wthp_callback *self)
{
  wth_object_delete((struct wth_object *)self);
}

static inline void
wthp_compositor_free(struct wthp_compositor *self)
{
  wth_object_delete((struct wth_object *)self);
}

struct wthp_surface *
wthp_compositor_create_surface (struct wthp_compositor * wthp_compositor);

struct wthp_region *
wthp_compositor_create_region (struct wthp_compositor * wthp_compositor);

static inline void
wthp_blob_factory_free(struct wthp_blob_factory *self)
{
  wth_object_delete((struct wth_object *)self);
}

struct wthp_buffer *
wthp_blob_factory_create_buffer (struct wthp_blob_factory * wthp_blob_factory, uint32_t data_sz, void * data, int32_t width, int32_t height, int32_t stride, uint32_t format);

static inline void
wthp_buffer_free(struct wthp_buffer *self)
{
  wth_object_delete((struct wth_object *)self);
}

void
wthp_buffer_destroy (struct wthp_buffer * wthp_buffer);

static inline void
wthp_surface_free(struct wthp_surface *self)
{
  wth_object_delete((struct wth_object *)self);
}

void
wthp_surface_destroy (struct wthp_surface * wthp_surface);

void
wthp_surface_attach (struct wthp_surface * wthp_surface, struct wthp_buffer * buffer, int32_t x, int32_t y);

void
wthp_surface_damage (struct wthp_surface * wthp_surface, int32_t x, int32_t y, int32_t width, int32_t height);

struct wthp_callback *
wthp_surface_frame (struct wthp_surface * wthp_surface);

void
wthp_surface_set_opaque_region (struct wthp_surface * wthp_surface, struct wthp_region * region);

void
wthp_surface_set_input_region (struct wthp_surface * wthp_surface, struct wthp_region * region);

void
wthp_surface_commit (struct wthp_surface * wthp_surface);

void
wthp_surface_set_buffer_transform (struct wthp_surface * wthp_surface, int32_t transform);

void
wthp_surface_set_buffer_scale (struct wthp_surface * wthp_surface, int32_t scale);

void
wthp_surface_damage_buffer (struct wthp_surface * wthp_surface, int32_t x, int32_t y, int32_t width, int32_t height);

static inline void
wthp_seat_free(struct wthp_seat *self)
{
  wth_object_delete((struct wth_object *)self);
}

struct wthp_pointer *
wthp_seat_get_pointer (struct wthp_seat * wthp_seat);

struct wthp_keyboard *
wthp_seat_get_keyboard (struct wthp_seat * wthp_seat);

struct wthp_touch *
wthp_seat_get_touch (struct wthp_seat * wthp_seat);

void
wthp_seat_release (struct wthp_seat * wthp_seat);

static inline void
wthp_pointer_free(struct wthp_pointer *self)
{
  wth_object_delete((struct wth_object *)self);
}

void
wthp_pointer_set_cursor (struct wthp_pointer * wthp_pointer, uint32_t serial, struct wthp_surface * surface, int32_t hotspot_x, int32_t hotspot_y);

void
wthp_pointer_release (struct wthp_pointer * wthp_pointer);

static inline void
wthp_keyboard_free(struct wthp_keyboard *self)
{
  wth_object_delete((struct wth_object *)self);
}

void
wthp_keyboard_release (struct wthp_keyboard * wthp_keyboard);

static inline void
wthp_touch_free(struct wthp_touch *self)
{
  wth_object_delete((struct wth_object *)self);
}

void
wthp_touch_release (struct wthp_touch * wthp_touch);

static inline void
wthp_output_free(struct wthp_output *self)
{
  wth_object_delete((struct wth_object *)self);
}

static inline void
wthp_region_free(struct wthp_region *self)
{
  wth_object_delete((struct wth_object *)self);
}

void
wthp_region_destroy (struct wthp_region * wthp_region);

void
wthp_region_add (struct wthp_region * wthp_region, int32_t x, int32_t y, int32_t width, int32_t height);

void
wthp_region_subtract (struct wthp_region * wthp_region, int32_t x, int32_t y, int32_t width, int32_t height);

static inline void
wthp_farstream_free(struct wthp_farstream *self)
{
  wth_object_delete((struct wth_object *)self);
}

void
wthp_farstream_destroy (struct wthp_farstream * wthp_farstream);

struct wthp_farstream_remote *
wthp_farstream_connect (struct wthp_farstream * wthp_farstream, uint32_t port);

static inline void
wthp_farstream_remote_free(struct wthp_farstream_remote *self)
{
  wth_object_delete((struct wth_object *)self);
}

void
wthp_farstream_remote_destroy (struct wthp_farstream_remote * wthp_farstream_remote);

void
wthp_farstream_remote_codec_offer (struct wthp_farstream_remote * wthp_farstream_remote, const char * list);

struct wthp_buffer *
wthp_farstream_remote_create_buffer (struct wthp_farstream_remote * wthp_farstream_remote, uint32_t frame_id);

#ifdef  __cplusplus
}
#endif

#endif